#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
In this module you find the base workflow for a impurity DOS calculation and
some helper methods to do so with AiiDA
"""

from aiida.orm import Code, load_node, CalcJobNode, Float, Int, Str, Dict, RemoteData, SinglefileData, XyData, Bool
from aiida.plugins import DataFactory
from aiida.engine import if_, ToContext, WorkChain, calcfunction
from aiida.common import LinkType
from aiida.common.folders import SandboxFolder
from aiida_kkr.tools import truncate_string
from aiida_kkr.workflows.gf_writeout import kkr_flex_wc
from aiida_kkr.workflows.kkr_imp_sub import kkr_imp_sub_wc
from aiida_kkr.workflows.dos import kkr_dos_wc
from aiida_kkr.calculations import KkrimpCalculation
import tarfile
import os
from aiida_kkr.tools.save_output_nodes import create_out_dict_node

__copyright__ = (u'Copyright (c), 2019, Forschungszentrum Jülich GmbH, '
                 'IAS-1/PGI-1, Germany. All rights reserved.')
__license__ = 'MIT license, see LICENSE.txt file'
__version__ = '0.7.2'
__contributors__ = (u'Fabian Bertoldo', u'Philipp Rüßmann')

# activate verbose output, for debugging only
_VERBOSE_ = True

#TODO: improve workflow output node structure
#TODO: generalise search for imp_info and conv_host from startpot


class kkr_imp_dos_wc(WorkChain):
    """
    Workchain of a DOS calculation for an impurity system starting from a
    converged impurity calculation or workflow

    :param options: (Dict), computer options
    :param wf_parameters: (Dict), specifications for the DOS
    :param kkr: (Code), mandatory: KKR code for gf_writeout step
    :param kkrimp: (Code), mandatory: KKRimp code for DOS calculation
    :param imp_host_pot: (SinglefileData), mandatory: impurity startpotential

    :return workflow_info: (Dict), Information on workflow results
    :return last_calc_output_parameters: (Dict), output parameters of
                                         the last called calculation
    :return last_calc_info: (Dict), information of the last called calculation
    """

    _workflowversion = __version__
    _wf_label = 'kkr_imp_dos_wc'
    _wf_description = 'Workflow for a KKR impurity DOS calculation'

    _options_default = {
        'queue_name': '',  # Queue name to submit jobs too
        'resources': {
            'num_machines': 1
        },  # resources to allocate for the job
        'max_wallclock_seconds': 60 * 60,  # walltime after which the job gets killed (gets parsed to KKR)}
        'custom_scheduler_commands': '',  # some additional scheduler commands
        'withmpi': True
    }  # execute KKR with mpi or without

    _wf_default = {
        'jij_run': False,  # calculate Jij's energy resolved
        'lmdos': False,  # calculate also (l,m) or only l-resolved DOS
        'retrieve_kkrflex': True,  # retrieve kkrflex files to repository or leave on remote computer only
    }
    # add defaults of dos_params since they are passed onto that workflow
    _wf_default['dos_params'] = kkr_dos_wc.get_wf_defaults(silent=True)

    @classmethod
    def get_wf_defaults(self):
        """
        Print and return _wf_defaults dictionary. Can be used to easily create set of wf_parameters.
        returns _wf_defaults
        """

        print(f'Version of workflow: {self._workflowversion}')
        return self._wf_default.copy()

    @classmethod
    def define(cls, spec):
        """
        Defines the outline of the workflow
        """

        # Take input of the workflow or use defaults defined above
        super(kkr_imp_dos_wc, cls).define(spec)

        spec.input('kkr', valid_type=Code, required=False, help='KKRhost code, needed if gf_dos_remote is not given.')
        spec.input('kkrimp', valid_type=Code, required=True, help='KKRimp code, always needed.')
        spec.input(
            'options',
            valid_type=Dict,
            required=False,
            default=lambda: Dict(dict=cls._options_default),
            help='Computer options (resources, quene name, etc.).'
        )
        spec.input(
            'wf_parameters',
            valid_type=Dict,
            required=False,
            default=lambda: Dict(dict=cls._wf_default),
            help='DOS workflow parameters (energy range, etc.).'
        )
        spec.input(
            'host_remote',
            valid_type=RemoteData,
            required=False,
            help='RemoteData node of the (converged) host calculation.'
        )
        spec.input(
            'gf_dos_remote',
            valid_type=RemoteData,
            required=False,
            help='RemoteData node of precomputed host GF for DOS energy contour.'
        )
        spec.input(
            'kkrimp_remote',
            valid_type=RemoteData,
            required=False,
            help='RemoteData node of previous (converged) KKRimp calculation.'
        )
        spec.input(
            'imp_pot_sfd',
            valid_type=SinglefileData,
            required=False,
            help='impurity potential single file data. Needs also impurity_info node.'
        )
        spec.input(
            'impurity_info',
            valid_type=Dict,
            required=False,
            help=
            'impurity info node that specifies the relation between imp_pot_sfd to the host system. Mandatory if imp_pot_sfd is given.'
        )
        spec.input(
            'params_kkr_overwrite',
            valid_type=Dict,
            required=False,
            help='Set some input parameters of the KKR calculation.'
        )
        spec.input(
            'settings_LDAU',
            valid_type=Dict,
            required=False,
            help='Settings for LDA+U run (see KkrimpCalculation for details).'
        )

        spec.expose_inputs(kkr_imp_sub_wc, namespace='BdG', include=('params_overwrite'))
        spec.expose_inputs(kkr_imp_sub_wc, include=('initial_noco_angles', 'rimpshift'))
        spec.expose_inputs(kkr_flex_wc, namespace='gf_writeout', include=('params_kkr_overwrite', 'options'))

        # specify the outputs
        spec.output('workflow_info', valid_type=Dict)
        spec.output('last_calc_output_parameters', valid_type=Dict)
        spec.output('last_calc_info', valid_type=Dict)
        spec.output('dos_data', valid_type=XyData)
        spec.output('dos_data_interpol', valid_type=XyData, required=False)
        spec.output('dos_data_lm', valid_type=XyData, required=False)
        spec.output('dos_data_interpol_lm', valid_type=XyData, required=False)
        spec.output(
            'gf_dos_remote', valid_type=RemoteData, required=False, help='RemoteData node of the computed host GF.'
        )

        # Here the structure of the workflow is defined
        spec.outline(
            cls.start,  # start and initialise workflow
            if_(cls.validate_input
                )(  # validate the given input
                    cls.run_gfstep,  # run GF step with DOS energy contour
                    cls.run_imp_dos
                ),  # run DOS for the impurity problem
            cls.return_results  # terminate workflow and return results
        )

        # Define possible exit codes for the workflow
        spec.exit_code(220, 'ERROR_UNKNOWN_PROBLEM', message='Unknown problem detected.')
        spec.exit_code(
            221,
            'ERROR_NO_PARENT_FOUND',
            message='Unable to find the parent remote_data node that led to '
            'the input impurity calculation. You need to specify '
            '`host_remote` and `impurity_info` nodes.'
        )
        spec.exit_code(
            222,
            'ERROR_GF_WRITEOUT_UNSUCCESFUL',
            message='The gf_writeout workflow was not succesful, cannot continue.'
        )
        spec.exit_code(
            223,
            'ERROR_IMP_POT_AND_REMOTE',
            message='The input nodes `imp_pot_sfd` and `kkrimp_remote` are given but are mutually exclusive'
        )
        spec.exit_code(
            224,
            'ERROR_KKR_CODE_MISSING',
            message='KKRhost code node (`inputs.kkr`) is missing if gf_dos_remote is not given.'
        )
        spec.exit_code(
            225, 'ERROR_HOST_REMOTE_MISSING', message='`host_remote` node is missing if gf_dos_remote is not given.'
        )
        spec.exit_code(226, 'ERROR_IMP_SUB_WORKFLOW_FAILURE', message='KKRimp sub-workflow failed.')

    def start(self):
        """
        Initialise context and some parameters
        """

        self.report(f'INFO: started KKR impurity DOS workflow version {self._workflowversion}')
        if _VERBOSE_:
            self.report(f'inputs: {self.inputs}')

        # input both wf and options parameters
        if 'wf_parameters' in self.inputs:
            wf_dict = self.inputs.wf_parameters.get_dict()
            if wf_dict == {}:
                wf_dict = self._wf_default
                self.report('INFO: using default wf parameters')

        if 'options' in self.inputs:
            options_dict = self.inputs.options.get_dict()
            if options_dict == {}:
                options_dict = self._options_default
                self.report('INFO: using default options parameters')

        # set values, or defaults
        self.ctx.withmpi = options_dict.get('withmpi', self._options_default['withmpi'])
        self.ctx.resources = options_dict.get('resources', self._options_default['resources'])
        self.ctx.max_wallclock_seconds = options_dict.get(
            'max_wallclock_seconds', self._options_default['max_wallclock_seconds']
        )
        self.ctx.queue = options_dict.get('queue_name', self._options_default['queue_name'])
        self.ctx.custom_scheduler_commands = options_dict.get(
            'custom_scheduler_commands', self._options_default['custom_scheduler_commands']
        )
        self.ctx.options_params_dict = Dict({
            'withmpi': self.ctx.withmpi,
            'resources': self.ctx.resources,
            'max_wallclock_seconds': self.ctx.max_wallclock_seconds,
            'queue_name': self.ctx.queue,
            'custom_scheduler_commands': self.ctx.custom_scheduler_commands
        })

        # set workflow parameters for the KKR imputrity calculations
        self.ctx.lmdos = wf_dict.get('lmdos', self._wf_default['lmdos'])
        self.ctx.retrieve_kkrflex = wf_dict.get('retrieve_kkrflex', self._wf_default['retrieve_kkrflex'])

        self.ctx.dos_params_dict = wf_dict.get('dos_params', self._wf_default['dos_params'])
        # fill missing key, value pairs with defaults
        for k, v in self._wf_default['dos_params'].items():
            if k not in self.ctx.dos_params_dict.keys():
                self.ctx.dos_params_dict[k] = v

        # set workflow parameters for the KKR impurity calculation
        self.ctx.jij_run = wf_dict.get('jij_run', self._wf_default['jij_run'])

        # set workflow label and description
        self.ctx.description_wf = self.inputs.get('description', self._wf_description)
        self.ctx.label_wf = self.inputs.get('label', self._wf_label)

        # whether or not to compute the GF writeout step
        self.ctx.skip_gfstep = False

        message = f"""
INFO: use the following parameter:
withmpi: {self.ctx.withmpi}
Resources: {self.ctx.resources}
Walltime (s): {self.ctx.max_wallclock_seconds}
queue name: {self.ctx.queue}
scheduler command: {self.ctx.custom_scheduler_commands}
description: {self.ctx.description_wf}
label: {self.ctx.label_wf}
                  """
        print(message)
        self.report(message)

        # return para/vars
        self.ctx.successful = True
        self.ctx.errors = []
        self.ctx.formula = ''

    def validate_input(self):
        """
        Validate input and catch possible errors
        """

        inputs = self.inputs
        inputs_ok = True
        gf_writeout_calc = None

        if 'imp_pot_sfd' in inputs:
            # check if input potential has incoming return link
            if len(inputs.imp_pot_sfd.get_incoming(link_type=LinkType.RETURN).all()) == 0:
                inputs_ok = self._imp_pot_not_from_wf(inputs_ok)
            else:
                gf_writeout_calc = self._imp_pot_from_wf()

        if 'gf_dos_remote' in self.inputs:
            self.ctx.skip_gfstep = True
        else:
            inputs_ok = self._check_gf_writeout_inputs(inputs_ok, gf_writeout_calc)

        if 'imp_pot_sfd' in self.inputs and 'kkrimp_remote' in self.inputs:
            self.report('[ERROR] both `imp_pot_sfd` and `kkrimp_remote` node in inputs')
            inputs_ok = False
            self.ctx.errors.append(2)  # raises ERROR_IMP_POT_AND_REMOTE
        elif 'imp_pot_sfd' in self.inputs:
            self.report('[INFO] use `imp_pot_sfd` input node')
        elif 'kkrimp_remote' in self.inputs:
            self.report('[INFO] use `kkrimp_remote` input node')
            # extract imp_info node from parent KKRimp calculation
            parent_impcalc = self.inputs.kkrimp_remote.get_incoming(node_class=CalcJobNode).first().node
            self.ctx.imp_info = parent_impcalc.inputs.impurity_info
        else:
            self.report('neither `imp_pot_sfd` nor `kkrimp_remote` node in inputs')
            inputs_ok = False
            self.ctx.errors.append(1)  # raises ERROR_NO_PARENT_FOUND

        message = f'INFO: validated input successfully: {inputs_ok}'
        self.report(message)

        return inputs_ok

    def _imp_pot_not_from_wf(self, inputs_ok):
        """
        input impurity potential is not from a kkrimp workflow
        this means we need also have the `impurity_info` and `host_remote` input nodes
        to be able to run the calculation
        """
        self.report('input potential not from kkrimp workflow: take remote_data folder of host system from input')
        inputs = self.inputs
        if 'impurity_info' in inputs and 'host_remote' in inputs:
            self.ctx.imp_info = inputs.impurity_info
            self.ctx.conv_host_remote = inputs.host_remote
        else:
            message = 'WARNING: startpot has no parent and can not find a converged host RemoteData node'
            self.report(message)
            if 'impurity_info' not in inputs:
                message = '`impurity_info` optional input node not given but needed in this case.'
                self.report(message)
            if 'host_remote' not in inputs:
                message = '`host_remote` optional input node not given but needed in this case.'
                self.report(message)
            inputs_ok = False
            self.ctx.errors.append(1)

        return inputs_ok

    def _imp_pot_from_wf(self):
        """
        we try to extract the input nodes for the GF writeout step from the incoming links to the impurity potential
        """
        # if return link is found get input nodes automatically
        message = 'INFO: get converged host RemoteData node and impurity_info node from database'
        self.report(message)
        self.ctx.kkr_imp_wf = self.inputs.imp_pot_sfd.get_incoming(node_class=kkr_imp_sub_wc).first().node
        message = f'INFO: found underlying kkr impurity workflow (pk: {self.ctx.kkr_imp_wf.pk})'
        self.report(message)
        self.ctx.imp_info = self.ctx.kkr_imp_wf.inputs.impurity_info
        message = f'INFO: found impurity_info node (pk: {self.ctx.imp_info.pk})'
        self.report(message)

        if 'remote_data' in self.ctx.kkr_imp_wf.inputs:
            remote_data_gf_writeout = self.ctx.kkr_imp_wf.inputs.remote_data
            gf_writeout_calc = remote_data_gf_writeout.get_incoming(node_class=CalcJobNode).first().node
            self.ctx.conv_host_remote = gf_writeout_calc.inputs.parent_folder
            message = 'INFO: imported converged_host_remote (pk: {}) and impurity_info from database'.format(
                self.ctx.conv_host_remote.pk
            )
            self.report(message)
        else:
            # follow links in DB upwards to get the remote_folder of the parent calculation
            self.ctx.conv_host_remote = self.ctx.kkr_imp_wf.inputs.gf_remote.inputs.remote_folder.inputs.parent_calc_folder.inputs.remote_folder.outputs.remote_folder
            message = 'INFO: imported converged_host_remote (pk: {}) and impurity_info from database'.format(
                self.ctx.conv_host_remote.pk
            )
            self.report(message)

        if 'host_remote' in self.inputs:
            self.ctx.conv_host_remote = self.inputs.host_remote

        return gf_writeout_calc

    def _check_gf_writeout_inputs(self, inputs_ok, gf_writeout_calc):
        """
        check if all necessary inputs are there for the GF writeout step
        """
        if 'kkr' not in self.inputs:
            self.report('[ERROR] `kkr` input node needed if `gf_dos_remote` is not given')
            inputs_ok = False
            self.ctx.errors.append(3)  # raises ERROR_KKR_CODE_MISSING

        if gf_writeout_calc is not None:
            self.report('Use extraced host remote')
        elif 'host_remote' not in self.inputs:
            self.report('[ERROR] `host_remote` input node needed if `gf_dos_remote` is not given')
            inputs_ok = False
            self.ctx.errors.append(4)  # raises ERROR_HOST_REMOTE_MISSING
        else:
            self.ctx.conv_host_remote = self.inputs.host_remote

        return inputs_ok

    def run_gfstep(self):
        """
        Start GF writeout step with DOS energy contour
        """

        if not self.ctx.skip_gfstep:
            options = self.ctx.options_params_dict
            kkrcode = self.inputs.kkr
            converged_host_remote = self.ctx.conv_host_remote
            imp_info = self.ctx.imp_info

            wf_params_gf = {
                'dos_run': True,
                'dos_params': self.ctx.dos_params_dict,
            }
            # maybe set kkrflex_retrieve
            if not self.ctx.retrieve_kkrflex:
                wf_params_gf['retrieve_kkrflex'] = self.ctx.retrieve_kkrflex
            # now convert to AiiDA Dict
            wf_params_gf = Dict(wf_params_gf)

            label_gf = 'GF writeout for imp DOS'
            description_gf = 'GF writeout step with energy contour for impurity DOS'

            builder = kkr_flex_wc.get_builder()
            builder.metadata.label = label_gf  # pylint: disable=no-member
            builder.metadata.description = description_gf  # pylint: disable=no-member
            builder.kkr = kkrcode
            builder.options = options
            builder.wf_parameters = wf_params_gf
            builder.remote_data = converged_host_remote
            builder.impurity_info = imp_info
            if 'params_kkr_overwrite' in self.inputs:
                builder.params_kkr_overwrite = self.inputs.params_kkr_overwrite
            if 'gf_writeout' in self.inputs:
                if 'params_kkr_overwrite' in self.inputs.gf_writeout:
                    builder.params_kkr_overwrite = self.inputs.gf_writeout.params_kkr_overwrite
                if 'options' in self.inputs.gf_writeout:
                    builder.options = self.inputs.gf_writeout.options

            future = self.submit(builder)

            message = f'INFO: running GF writeout (pid: {future.pk})'
            self.report(message)

            return ToContext(gf_writeout=future)

    def run_imp_dos(self):
        """
        Use previous GF step to calculate DOS for the impurity problem
        """

        if not self.ctx.skip_gfstep:
            # use computed gf_writeout
            if not self.ctx.gf_writeout.is_finished_ok:
                return self.exit_codes.ERROR_GF_WRITEOUT_UNSUCCESFUL  # pylint: disable=no-member
            gf_writeout_wf = self.ctx.gf_writeout
            gf_writeout_calc = load_node(self.ctx.gf_writeout.outputs.workflow_info.get_dict().get('pk_flexcalc'))
            gf_writeout_remote = gf_writeout_wf.outputs.GF_host_remote
            self.ctx.pk_flexcalc = self.ctx.gf_writeout.outputs.workflow_info.get_dict().get('pk_flexcalc')
        else:
            # use gf_writeout from input
            gf_writeout_remote = self.inputs.gf_dos_remote
            gf_writeout_calc = gf_writeout_remote.get_incoming(node_class=CalcJobNode).first().node
            self.ctx.pk_flexcalc = gf_writeout_calc.pk

        self.ctx.gf_writeout_remote = gf_writeout_remote

        options = self.ctx.options_params_dict
        kkrimpcode = self.inputs.kkrimp
        if 'imp_pot_sfd' in self.inputs:
            # take impurit potential SingleFileData node
            impurity_pot_or_remote = self.inputs.imp_pot_sfd
        elif 'kkrimp_remote' in self.inputs:
            # or from RemoteData node of previous KKRimp calc
            impurity_pot_or_remote = self.inputs.kkrimp_remote
        imps = self.ctx.imp_info
        nspin = gf_writeout_calc.outputs.output_parameters.get_dict().get('nspin')
        self.ctx.nspin = nspin
        self.report(f'nspin: {nspin}')
        self.ctx.kkrimp_params_dict = Dict({
            'nspin': nspin,
            'nsteps': 1,
            'kkr_runmax': 1,
            'dos_run': True,
            'lmdos': self.ctx.lmdos,
            'jij_run': self.ctx.jij_run,
        })
        kkrimp_params = self.ctx.kkrimp_params_dict
        label_imp = 'KKRimp DOS (GF: {}, imp_pot: {}, Zimp: {}, ilayer_cent: {})'.format(
            gf_writeout_calc.pk,
            impurity_pot_or_remote.pk,  # pylint: disable=possibly-used-before-assignment
            imps.get_dict().get('Zimp'),
            imps.get_dict().get('ilayer_center')
        )
        description_imp = 'KKRimp DOS run (GF: {}, imp_pot: {}, Zimp: {}, ilayer_cent: {}, R_cut: {})'.format(
            gf_writeout_calc.pk, impurity_pot_or_remote.pk,
            imps.get_dict().get('Zimp'),
            imps.get_dict().get('ilayer_center'),
            imps.get_dict().get('Rcut')
        )

        builder = kkr_imp_sub_wc.get_builder()
        # Attention: label and description must not be loo long (255 characters max)!
        builder.metadata.label = truncate_string(label_imp, 255)  # pylint: disable=no-member
        builder.metadata.description = description_imp  # pylint: disable=no-member
        builder.kkrimp = kkrimpcode
        builder.options = options
        builder.wf_parameters = kkrimp_params
        builder.remote_data = gf_writeout_remote
        if 'imp_pot_sfd' in self.inputs:
            self.report('Using impurity potential SingelfilData as input')
            builder.host_imp_startpot = impurity_pot_or_remote
        else:
            self.report('Using KKRimp remote folder as input')
            builder.kkrimp_remote = impurity_pot_or_remote
        builder.impurity_info = imps
        # LDA+U settings
        if 'settings_LDAU' in self.inputs:
            self.report('Add settings_LDAU input node')
            builder.settings_LDAU = self.inputs.settings_LDAU

        if 'params_overwrite' in self.inputs.BdG:
            builder.params_overwrite = self.inputs.BdG.params_overwrite
        if 'initial_noco_angles' in self.inputs:
            builder.initial_noco_angles = self.inputs.initial_noco_angles
        if 'rimpshift' in self.inputs:
            builder.rimpshift = self.inputs.rimpshift

        future = self.submit(builder)

        message = f'INFO: running DOS step for impurity system (pk: {future.pk})'
        print(message)
        self.report(message)

        return ToContext(kkrimp_dos=future)

    def return_results(self):
        """
        Return the results and create all of the output nodes
        """

        if self.ctx.errors != []:
            if 1 in self.ctx.errors:
                return self.exit_codes.ERROR_NO_PARENT_FOUND  # pylint: disable=no-member
            elif 2 in self.ctx.errors:
                return self.exit_codes.ERROR_IMP_POT_AND_REMOTE  # pylint: disable=no-member
            elif 3 in self.ctx.errors:
                return self.exit_codes.ERROR_KKR_CODE_MISSING  # pylint: disable=no-member
            elif 4 in self.ctx.errors:
                return self.exit_codes.ERROR_HOST_REMOTE_MISSING  # pylint: disable=no-member
            else:
                return self.exit_codes.ERROR_UNKNOWN_PROBLEM  # pylint: disable=no-member

        message = 'INFO: creating output nodes for the KKR imp DOS workflow ...'
        print(message)
        self.report(message)

        if not self.ctx.kkrimp_dos.is_finished_ok:
            message = 'ERROR: sub workflow for impurity calculation failed'
            print(message)
            self.report(message)
            return self.exit_codes.ERROR_IMP_SUB_WORKFLOW_FAILURE  # pylint: disable=no-member
        else:
            last_calc_pk = self.ctx.kkrimp_dos.outputs.workflow_info.get_dict().get('last_calc_nodeinfo')['pk']
            last_calc = load_node(last_calc_pk)
            last_calc_output_params = last_calc.outputs.output_parameters
            last_calc_info = self.ctx.kkrimp_dos.outputs.workflow_info

            outputnode_dict = {}
            outputnode_dict['impurity_info'] = self.ctx.imp_info.get_dict()
            outputnode_dict['workflow_name'] = self.__class__.__name__
            outputnode_dict['workflow_version'] = self._workflowversion
            if not self.ctx.skip_gfstep:
                outputnode_dict['used_subworkflows'] = {'gf_writeout': self.ctx.gf_writeout.pk}
            else:
                outputnode_dict['used_subworkflows'] = {}
            outputnode_dict['used_subworkflows']['impurity_dos'] = self.ctx.kkrimp_dos.pk

            # interpol dos file and store to XyData nodes
            dos_extracted, dosXyDatas = self.extract_dos_data(last_calc)
            message = f'INFO: extracted DOS data? {dos_extracted}'
            print(message)
            self.report(message)

            # create results node and link rest of results
            link_nodes = dosXyDatas.copy()
            link_nodes['last_calc_output_parameters'] = last_calc_output_params
            link_nodes['last_calc_remote'] = last_calc.outputs.remote_folder
            outputnode_t = create_out_dict_node(Dict(dict=outputnode_dict), **link_nodes)
            outputnode_t.label = 'kkr_imp_dos_wc_inform'
            outputnode_t.description = 'Contains information for workflow'

            if dos_extracted:
                self.out('dos_data', dosXyDatas['dos_data'])
                if 'dos_data_interpol' in dosXyDatas:
                    self.out('dos_data_interpol', dosXyDatas['dos_data_interpol'])
                if self.ctx.lmdos:
                    self.out('dos_data_lm', dosXyDatas['dos_data_lm'])
                    if 'dos_data_interpol_lm' in dosXyDatas:
                        self.out('dos_data_interpol_lm', dosXyDatas['dos_data_interpol_lm'])

            message = f'INFO: workflow_info node: {outputnode_t.uuid}'
            print(message)
            self.report(message)

            self.out('workflow_info', outputnode_t)
            self.out('last_calc_output_parameters', last_calc_output_params)
            self.out('last_calc_info', last_calc_info)
            self.out('gf_dos_remote', self.ctx.gf_writeout_remote)

            message = 'INFO: created output nodes for KKR imp DOS workflow.'
            print(message)
            self.report(message)
            self.report(
                '\n'
                '|------------------------------------------------------------------------------------------------------------------|\n'
                '|-------------------------------------| Done with the KKR imp DOS workflow! |--------------------------------------|\n'
                '|------------------------------------------------------------------------------------------------------------------|'
            )

    def extract_dos_data(self, last_calc):
        """
        Extract DOS data from retrieved folder of KKRimp calculation.
        If output is compressed in tarfile take care of extracting this before parsing `out_ldos*` files.

        The extraction of the DOS data is done in `self.extract_dos_data_from_folder()` calls.

        returns:
          :boolean dos_extracted: logical signalling if extraction of DOS files was successful
          :dictionary dosXyDatas: dictionary containing dos data and interpolated dos data
        """

        # here we look for the dos files or the tarball containing the dos files:
        dos_retrieved = last_calc.outputs.retrieved

        if KkrimpCalculation._FILENAME_TAR in dos_retrieved.list_object_names():
            # deal with packed output files (overwrites dos_retrieved with sandbox into which files are extracted
            # this way the unpacked files are deleted after parsing and only the tarball is kept in the retrieved directory

            # for this we create a Sandbox folder
            with SandboxFolder() as tempfolder:
                # get abs paths of tempfolder and tarfile (needed by extract method of tarfile)

                # get abs path of tempfolder
                with tempfolder.open('.dummy', 'w') as tmpfile:
                    tempfolder_path = os.path.dirname(tmpfile.name)
                # get path of tarfile
                with dos_retrieved.open(KkrimpCalculation._FILENAME_TAR) as tf:
                    tfpath = tf.name

                # extract file from tarfile of retrieved to tempfolder
                with tarfile.open(tfpath) as tf:
                    tar_filenames = [ifile.name for ifile in tf.getmembers()]
                    for filename in tar_filenames:
                        if 'dos' in filename:  # should extract all out_ldos*, out_lmdos* files
                            tf.extract(filename, tempfolder_path)  # extract to tempfolder

                # now files are in tempfolder from where we can extract the dos data
                dos_extracted, dosXyDatas = self.extract_dos_data_from_folder(tempfolder, last_calc)
        else:
            # extract directly from retrieved (no tarball there)
            dos_extracted, dosXyDatas = self.extract_dos_data_from_folder(dos_retrieved, last_calc)

        return dos_extracted, dosXyDatas

    def extract_dos_data_from_folder(self, folder, last_calc):
        """
        Get DOS data and parse files.
        """

        # initialize in case dos data is not extracted
        dosXyDatas = None

        # get list of files in directory (needed since SandboxFolder does not have `list_object_names` method)
        # also extract absolute path of folder (needed by parse_impdosfiles since calcfunction does not work with SandboxFolder as input)
        if isinstance(folder, SandboxFolder):
            folder_abspath = folder.abspath
            filelist = os.listdir(folder_abspath)
        else:
            filelist = folder.list_object_names()

        # check if out_ldos* files are there and parse dos files
        if 'out_ldos.interpol.atom=01_spin1.dat' in filelist or 'out_ldos.interpol.atom=001_spin1.dat' in filelist:
            # extract EF and number of atoms from kkrflex_writeout calculation
            kkrflex_writeout = load_node(self.ctx.pk_flexcalc)
            parent_calc_kkr_converged = kkrflex_writeout.inputs.parent_folder.get_incoming(node_class=CalcJobNode
                                                                                           ).first().node
            ef = parent_calc_kkr_converged.outputs.output_parameters.get_dict().get('fermi_energy')
            last_calc_output_params = last_calc.outputs.output_parameters
            natom = last_calc_output_params.get_dict().get('number_of_atoms_in_unit_cell')
            # parse dosfiles using nspin, EF and Natom inputs
            dosXyDatas = parse_impdosfiles(folder, Int(natom), Int(self.ctx.nspin), Float(ef), Bool(False))
            if self.ctx.lmdos:
                dosXyDatas2 = parse_impdosfiles(folder, Int(natom), Int(self.ctx.nspin), Float(ef), Bool(True))
                dosXyDatas['dos_data_lm'] = dosXyDatas2['dos_data']
                dosXyDatas['dos_data_interpol_lm'] = dosXyDatas2['dos_data_interpol']
            dos_extracted = True
        else:
            dos_extracted = False
            dosXyDatas = None

        return dos_extracted, dosXyDatas


@calcfunction
def parse_impdosfiles(folder, natom, nspin, ef, use_lmdos):
    """
    Read `out_ldos*` files and create XyData node with l-resolved DOS (+node for interpolated DOS if files are found)

    Inputs:
    :param dos_abspath: absolute path to folder where `out_ldos*` files reside (AiiDA Str object)
    :param natom: number of atoms (AiiDA Int object)
    :param nspin: number of spin channels (AiiDA Int object)
    :param ef: Fermi energy in Ry units (AiiDA Float object)

    Returns:
    output dictionary containing
      output = {'dos_data': dosnode, 'dos_data_interpol': dosnode2}
    where `dosnode` and `dosnode2` are AiiDA XyData objects
    """
    from masci_tools.io.common_functions import get_Ry2eV, get_ef_from_potfile
    from numpy import loadtxt, array

    # read dos files
    dos, dos_int = [], []
    for iatom in range(1, natom.value + 1):
        for ispin in range(1, nspin.value + 1):
            name0 = 'out_ldos'
            if use_lmdos.value:
                name0 = 'out_lmdos'
            fname = name0 + '.atom=%0.2i_spin%i.dat' % (iatom, ispin)
            if fname in folder.list_object_names():
                # old file names with 2 digits for atom index
                with folder.open(fname) as dosfile:
                    tmp = loadtxt(dosfile)
                    dos.append(tmp)

                with folder.open(name0 + '.interpol.atom=%0.2i_spin%i.dat' % (iatom, ispin)) as dosfile:
                    tmp = loadtxt(dosfile)
                    if len(tmp) > 0:
                        dos_int.append(tmp)
            else:
                # new file names with 3 digits for atom numbers
                with folder.open(name0 + '.atom=%0.3i_spin%i.dat' % (iatom, ispin)) as dosfile:
                    tmp = loadtxt(dosfile)
                    dos.append(tmp)

                with folder.open(name0 + '.interpol.atom=%0.3i_spin%i.dat' % (iatom, ispin)) as dosfile:
                    try:
                        tmp = loadtxt(dosfile)
                    except:
                        tmp = []
                    # can happen if there are too few points to interpolate on
                    if len(tmp) > 0:
                        dos_int.append(tmp)

    dos = array(dos)
    if len(dos_int) > 0:
        dos_int = array(dos_int)
    else:
        pass

    # convert to eV units
    eVscale = get_Ry2eV()
    dos[:, :, 0] = (dos[:, :, 0] - ef.value) * eVscale
    dos[:, :, 1:] = dos[:, :, 1:] / eVscale
    if len(dos_int) > 0:
        try:
            dos_int[:, :, 0] = (dos_int[:, :, 0] - ef.value) * eVscale
            dos_int[:, :, 1:] = dos_int[:, :, 1:] / eVscale
        except:
            message = 'Not enough data were present in the interpolated dos, due to a lack of energy points to interpolate on'
    else:
        pass

    # create output nodes
    dosnode = XyData()
    dosnode.label = 'dos_data'
    dosnode.description = 'Array data containing uniterpolated DOS (i.e. dos at finite imaginary part of energy). 3D array with (atoms, energy point, l-channel) dimensions.'
    dosnode.set_x(dos[:, :, 0], 'E-EF', 'eV')

    name = ['tot', 's', 'p', 'd', 'f', 'g']
    name = name[:len(dos[0, 0, 1:]) - 1] + ['ns']

    if use_lmdos.value:
        name = [
            's', 'p1', 'p2', 'p3', 'd1', 'd2', 'd3', 'd4', 'd5', 'f1', 'f2', 'f3', 'f4', 'f5', 'f6', 'f7', 'g1', 'g2',
            'g3', 'g4', 'g5', 'g6', 'g7', 'g8', 'g9'
        ]
        name = name[:len(dos[0, 0, 1:])]

    ylists = [[], [], []]
    for l in range(len(name)):
        ylists[0].append(dos[:, :, 1 + l])
        ylists[1].append('dos ' + name[l])
        ylists[2].append('states/eV')
    dosnode.set_y(ylists[0], ylists[1], ylists[2])

    # node for interpolated DOS
    try:
        dosnode2 = XyData()
        dosnode2.label = 'dos_interpol_data'
        dosnode2.description = 'Array data containing iterpolated DOS (i.e. dos at finite imaginary part of energy). 3D array with (atoms, energy point, l-channel) dimensions.'
        dosnode2.set_x(dos_int[:, :, 0], 'E-EF', 'eV')
        ylists = [[], [], []]
        for l in range(len(name)):
            ylists[0].append(dos_int[:, :, 1 + l])
            ylists[1].append('interpolated dos ' + name[l])
            ylists[2].append('states/eV')
        dosnode2.set_y(ylists[0], ylists[1], ylists[2])

        output = {'dos_data': dosnode, 'dos_data_interpol': dosnode2}
    except:
        output = {'dos_data': dosnode}

    return output



# -*- coding: utf-8 -*-
'''
Module with CLI commands to launch for calcjob and workflows of aiida-kkr.
'''
import click

from aiida.cmdline.params import types
from aiida.cmdline.params.options import OverridableOption
from aiida.orm import Code, load_node, Dict
from aiida.plugins import WorkflowFactory
from aiida.plugins import CalculationFactory
from masci_tools.io.kkr_params import kkrparams

from aiida_kkr.tools.dict_util import clean_nones
from ..util import options
from ..util.utils import launch_process
from ..util import defaults

# TODO: command for kkrimporter
# not sure if this should be a launch command
# TODO: commands for workchains: voro_start, eos, gf_writeout, kkr_imp, kkr_imp_dos, kkr_imp_sub
# Check_para_convergence, check_magnetic_state. base_restart_calc?


@click.command('voro')
@options.STRUCTURE_OR_FILE(default=defaults.get_cu_bulk_structure, show_default=True)
@options.VORO()
@options.PARAMETERS()
@options.PARENT_FOLDER()
@options.POTENTIAL_OVERWRITE()
@options.MAX_WALLCLOCK_SECONDS()
@options.QUEUE_NAME()
@options.DAEMON()
def launch_voro(
    structure,
    voro,
    parameters,
    parent_folder,
    potential_overwrite,
    max_wallclock_seconds,
    queue_name,
    daemon,
):
    """
    Launch an Voronoi calcjob on given input
    """
    # TODO?: maybe allow for additional metadata to be given.
    process_class = CalculationFactory('kkr.voro')

    inputs = {
        'structure': structure,
        'code': voro,
        'parameters': parameters,
        'parent_kkr': parent_folder,
        'potential_overwrite': potential_overwrite,
        'metadata': {
            'options': {
                'withmpi': False,
                'queue_name': queue_name,
                'max_wallclock_seconds': max_wallclock_seconds,
                'resources': {
                    'num_machines': 1,
                    'num_mpiprocs_per_machine': 1,
                }
            }
        }
    }
    inputs = clean_nones(inputs)
    builder = process_class.get_builder()
    builder.update(inputs)
    launch_process(builder, daemon)


@click.command('kkr')
@options.KKR()
@options.PARAMETERS()
@options.PARENT_FOLDER(required=True)
@options.IMPURITY_INFO()
@options.KPOINTS()
@options.DAEMON()
@options.WITH_MPI()
@options.NUM_MPIPROCS_PER_MACHINE()
@options.MAX_WALLCLOCK_SECONDS()
@options.MAX_NUM_MACHINES()
@options.QUEUE_NAME()
def launch_kkr(
    kkr,
    parameters,
    parent_folder,
    impurity_info,
    kpoints,
    daemon,
    with_mpi,
    num_mpiprocs_per_machine,
    max_wallclock_seconds,
    max_num_machines,
    queue_name,
):
    """
    Launch an KKRhost calcjob on given input
    """
    # TODO?: maybe allow for additional metadata to be given.
    process_class = CalculationFactory('kkr.kkr')

    inputs = {
        'code': kkr,
        'parameters': parameters,
        'parent_folder': parent_folder,
        'impurity_info': impurity_info,
        'metadata': {
            'options': {
                'withmpi': with_mpi,
                'max_wallclock_seconds': max_wallclock_seconds,
                'queue_name': queue_name,
                'resources': {
                    'num_machines': max_num_machines,
                    'num_mpiprocs_per_machine': num_mpiprocs_per_machine,
                },
            }
        }
    }
    inputs = clean_nones(inputs)
    builder = process_class.get_builder()
    builder.update(inputs)
    launch_process(builder, daemon)


@click.command('kkrimp')
@options.KKRIMP()
@options.PARAMETERS()
@options.PARENT_FOLDER(required=True)
@options.IMPURITY_INFO()
@options.DAEMON()
@options.WITH_MPI()
@options.NUM_MPIPROCS_PER_MACHINE()
@options.MAX_WALLCLOCK_SECONDS()
@options.MAX_NUM_MACHINES()
def launch_kkrimp(
    kkrimp,
    parameters,
    parent_folder,
    impurity_info,
    daemon,
    with_mpi,
    num_mpiprocs_per_machine,
    max_wallclock_seconds,
    max_num_machines,
):
    """
    Launch an KKRimp calcjob on given input
    """
    # TODO?: maybe allow for additional metadata to be given.
    process_class = CalculationFactory('kkr.kkrimp')

    inputs = {
        'code': kkrimp,
        'parameters': parameters,
        'parent_folder': parent_folder,
        'impurity_info': impurity_info,
        'metadata': {
            'options': {
                'withmpi': with_mpi,
                'max_wallclock_seconds': max_wallclock_seconds,
                'resources': {
                    'num_machines': max_num_machines,
                    'num_mpiprocs_per_machine': num_mpiprocs_per_machine,
                }
            }
        }
    }
    inputs = clean_nones(inputs)
    builder = process_class.get_builder()
    builder.update(inputs)
    launch_process(builder, daemon)


@click.command('dos')
@options.KKR()
@options.WF_PARAMETERS()
@options.OPTION_NODE()
@options.PARENT_FOLDER()
@options.DAEMON()
def launch_dos(kkr, wf_parameters, option_node, parent_folder, daemon):
    """
    Launch an KKRhost density of states workflow
    """
    # TODO?: maybe allow for additional metadata to be given.
    process_class = WorkflowFactory('kkr.dos')

    inputs = {
        'kkr': kkr,
        'remote_data': parent_folder,
        'options': option_node,
        'wf_parameters': wf_parameters,
    }
    inputs = clean_nones(inputs)
    builder = process_class.get_builder()
    builder.update(inputs)
    launch_process(builder, daemon)


@click.command('bs')
@options.KKR()
@options.WF_PARAMETERS()
@options.KPOINTS()
@options.OPTION_NODE()
@options.PARENT_FOLDER()
@options.DAEMON()
def launch_bs(kkr, wf_parameters, kpoints, option_node, parent_folder, daemon):
    """
    Launch an KKRhost bandstructure workflow with required inputs
    (kkr code, remote_data, options, wf-parameters, daemon)
    """
    process_class = WorkflowFactory('kkr.bs')

    inputs = {
        'kkr': kkr,
        'remote_data': parent_folder,
        'kpoints': kpoints,
        'options': option_node,
        'wf_parameters': wf_parameters,
    }
    inputs = clean_nones(inputs)
    builder = process_class.get_builder()
    builder.update(inputs)
    launch_process(builder, daemon)


@click.command('scf')
@options.KKR()
@options.VORO()
@options.STRUCTURE_OR_FILE(default=defaults.get_cu_bulk_structure, show_default=True)
@options.PARAMETERS()
@options.PARENT_FOLDER()
@options.DAEMON()
@options.OPTION_NODE()
@options.WF_PARAMETERS()
@options.POTENTIAL_OVERWRITE()
@options.NOCO_ANGLES()
def launch_scf(
    kkr,
    voro,
    structure,
    parameters,
    parent_folder,
    daemon,
    option_node,
    wf_parameters,
    potential_overwrite,
    noco_angles,
):
    """
    Launch an KKRhost self-consistency workflow
    """
    # TODO?: maybe allow for additional metadata to be given.
    process_class = WorkflowFactory('kkr.scf')

    inputs = {
        'kkr': kkr,
        'voronoi': voro,
        'structure': structure,
        'calc_parameters': parameters,
        'remote_data': parent_folder,
        'options': option_node,
        'wf_parameters': wf_parameters,
        'startpot_overwrite': potential_overwrite,
        'initial_noco_angles': noco_angles,
    }
    inputs = clean_nones(inputs)
    builder = process_class.get_builder()
    builder.update(inputs)
    launch_process(builder, daemon)


PARENT_FOR_IMP = OverridableOption(
    '-P',
    '--parent-folder',
    'parent_folder',
    type=types.DataParamType(sub_classes=('aiida.data:core.remote',)),
    show_default=True,
    required=True,
    help=
    'The remote folder of a parent calculation (either the converged calculation or a GF writeout calculation which prevents recalculating the host Greens function).'
)

PARAMS_HOST_GF = OverridableOption(
    '-p',
    '--parameters-hostGF',
    'parameters_hostgf',
    type=types.DataParamType(sub_classes=('aiida.data:core.dict',)),
    show_default=True,
    help='Set of parameters that are overwritten in the host GF writeout step.'
)

IMP_STARTPOT = OverridableOption(
    '--impurity-startpot',
    'impurity_startpot',
    type=types.DataParamType(sub_classes=('aiida.data:core.singlefile',)),
    help=
    'Use this as the starting potential for the impurity calculation. Needs to match the settings in the impurity info node.'
)


def _check_parent_calc_type(parent_folder):
    """
    check if parent folder was a host GF writeout calculation
    """
    from aiida_kkr.calculations import KkrCalculation
    calc = parent_folder.get_incoming(node_class=KkrCalculation).first()
    calc = calc.node
    ret = calc.outputs.retrieved
    return 'kkrflex_green' in ret.list_object_names()


@click.command('kkrimpscf')
@options.KKR()
@options.VORO()
@options.KKRIMP()
@options.IMPURITY_INFO(required=True, show_default=True)
@PARENT_FOR_IMP()
@options.OPTION_NODE()
@options.DAEMON()
@PARAMS_HOST_GF()
@options.WF_PARAMETERS()
@IMP_STARTPOT()
def launch_kkrimp_scf(
    kkr,
    voro,
    kkr_imp,
    impurity_info,
    parent_folder,
    option_node,
    daemon,
    parameters_hostgf,
    wf_parameters,
    impurity_startpot,
):
    """
    Launch an kkr calcjob on given input
    """
    # TODO?: maybe allow for additional metadata to be given.
    process_class = WorkflowFactory('kkr.imp')

    inputs = {
        'kkr': kkr,
        'kkrimp': kkr_imp,
        'voronoi': voro,
        'options': option_node,
        'wf_parameters': wf_parameters,
        'impurity_info': impurity_info,
        'params_kkr_overwrite': parameters_hostgf,
        'startpot': impurity_startpot,
    }
    inputs = clean_nones(inputs)

    if _check_parent_calc_type(parent_folder):
        # this means we can reuse the GF writeout from the parent calculation
        inputs['remote_data_gf'] = parent_folder
        click.echo('The parent_folder already has the written out GF of the host, we can therefore skip that step')
    else:
        # this means the calculation was not a GF writeout KKR calculation
        inputs['remote_data_host'] = parent_folder
        click.echo('The parent_folder does not contain the host GF, we calculate it.')

    builder = process_class.get_builder()
    builder.update(inputs)
    launch_process(builder, daemon)



# -*- coding: utf-8 -*-
"""
This module contains the workflow which combines pre-converged two single-impurity calculations to a larger impurity calculation
"""

from aiida.engine import WorkChain, if_, ToContext, calcfunction
from aiida.orm import load_node, Dict, WorkChainNode, Int, RemoteData, Bool, ArrayData, CalcJobNode
from aiida_kkr.calculations import KkrCalculation, KkrimpCalculation
from aiida_kkr.workflows import kkr_imp_sub_wc, kkr_flex_wc, kkr_imp_wc
from aiida_kkr.tools.combine_imps import (
    create_combined_imp_info_cf, combine_potentials_cf, get_zimp, get_host_structure, get_nspin_and_pot,
    combine_settings_ldau
)
from aiida_kkr.tools.save_output_nodes import create_out_dict_node
import tarfile
import numpy as np
from masci_tools.io.common_functions import get_Ry2eV

__copyright__ = (u'Copyright (c), 2020, Forschungszentrum Jülich GmbH, '
                 'IAS-1/PGI-1, Germany. All rights reserved.')
__license__ = 'MIT license, see LICENSE.txt file'
__version__ = '0.3.6'
__contributors__ = (u'Philipp Rüßmann , Rubel Mozumder, David Antognini Silva')

# activate debug writeout
_debug = True


class combine_imps_wc(WorkChain):
    """
    Workchain that combines 2 converged single-impurity calculations to a bigger impurity,
    reusing the preconverged potentials. This is useful, for example, to study co-doping.

   Inputs:
    :param impurity1_output_node:(Dict), required, output node from singel impurity wc, and should be one of the
                                 following:
                                 * for `kkr_imp_wc`: single_imp_worlfow.outputs.workflow_info
                                 * for `kkr_imp_sub_wc`: single_imp_worlfow.outputs.workflow_info
                                 * for `KkrimpCalculation`: single_imp_worlfow.outputs.output_parameters

    :param impurity2_output_node:(Dict), required, output node from second singel impurity wc, and should be one of
                                 the following:
                                 * for `kkr_imp_wc`: single_imp_worlfow.outputs.workflow_info
                                 * for `kkr_imp_sub_wc`: single_imp_worlfow.outputs.workflow_info
                                 * for `KkrimpCalculation`: single_imp_worlfow.outputs.output_parameters
    :offset_imp2:(Dict), required, offset of the second impurity with respect to the first impurity position.
                 e.g. {'index:0 or 1}, the replacement by the second impurity will take place at the same cell
                        or at the next nearest cell respectively.

    :param scf.kkrimp: (Code), mandatory, KKRimp code needed to submit kkr_imp_wc
    :param scf.wf_parameters: (Dict), optional, KKRimp code needed to submit kkr_imp_sub_wc
    :param scf.options: (Dict), optional, computer options for kkr_imp_sub_wc

    :param host_gf.kkr: (Code), optional, KKR code for submit kkr_flex_wc, needed if remote_data_gf is not given
    :param host_gf.options: (Dict), optional, computer options for kkr_flex_wc
    :param host_gf.params_kkr_overwrite: (Dict), optional, needed for kkr calculation for GF writeout

    :param wf_parameters_overwrite: (Dict), optional, specifications for wf_parameters of kkr_imp_sub_wc as well
                                  as well as wf_parameters of kkr_flex_wc.
    :param gf_host_remote: (RemoteData), optional, remote folder of a previous kkrflex writeout step
                          calculations containing the flexfiles and will be used for combine host GF.

   Returns:
    :return workflow_info: (Dict), Information of workflow results
    :return last_calc_output_parameters: (Dict), link to output parameters of the last called calculation of the
                                        scf kkr_imp_sub_wc.
    :return last_potential: (SingleFileData) link to last output potential of scf kkr_imp_sub_wc step.
    :return last_calc_remote: (RemoteData) link to remote data of last called calculation of the scf step.
    :return remote_data_gf: (RemoteData) link to GF_host_remote of outputs of kkr_flex_wc e.g. gf_writeou
                           step (only present of host GF was generated here).
    :return JijData: (ArrayData) Consists magnetic interaction data among the magnetic impurity atoms,
                    such as vector distance(rx, ry, rz) between atoms, spin interaction magnetude J,
                    Dzyaloshinskii-Moriya vector magnitude, and Dzyaloshinskii-Moriya vector component(Dx, Dy, Dz)
    :return JijInfo :(Dict) Consists description about the JijData.

    """

    _workflowversion = __version__
    _wf_default = {
        'jij_run': False,  # Any kind of addition in _wf_default should be updated into the start() as well.
        'allow_unconverged_inputs': False,  # start calculation only if previous impurity calculations are converged?
    }

    @classmethod
    def get_wf_defaults(cls, silent=False):
        """
        Print and return _wf_defaults dictionary. Can be used to easily create
        set of wf_parameters_overwrite.
        returns _wf_defaults
        """
        if not silent:
            print(f'Version of workflow: {cls._workflowversion}')
        return {
            'global': cls._wf_default.copy(),
            'scf': kkr_imp_sub_wc.get_wf_defaults(),
            'ghost_gf': kkr_flex_wc.get_wf_defaults()
        }

    @classmethod
    def define(cls, spec):
        """
        Defines the outline of the workflow
        """

        # take define from AiiDA base class and extend it then
        super(combine_imps_wc, cls).define(spec)

        # expose these inputs from sub-workflows
        spec.expose_inputs(
            kkr_imp_sub_wc,
            namespace='scf',
            include=('kkrimp', 'options', 'wf_parameters', 'settings_LDAU', 'params_overwrite')
        )
        spec.expose_inputs(
            kkr_flex_wc,
            namespace='host_gf',
            include=(
                'kkr',
                'options',
                'params_kkr_overwrite',
                'wf_parameters',
            ),  # expose only those port which are not set automatically
            namespace_options={
                'required': False,
                'populate_defaults': False
            },  # this makes sure the kkr code input is not needed if gf_host_remote is provided and the entire namespace is omitted
        )

        # mandatory inputs
        spec.input(
            'impurity1_output_node',
            required=True,
            valid_type=Dict,  #TODO make validator for input node to make sure it is the output of kkr_imp_wc
            help="""
Output node of a single impurity calculation. This can be the output of either the `kkr_imp_wc`, `kkr_imp_sub_wc`
workflows or of an `KkrimpCalculation`.

Use these output Dict nodes:
  * for `kkr_imp_wc`: single_imp_workfow.outputs.workflow_info
  * for `kkr_imp_sub_wc`: single_imp_workfow.outputs.workflow_info
  * for `KkrimpCalculation`: single_imp_workfow.outputs.output_parameters
"""
        )

        spec.input(
            'impurity2_output_node',
            required=True,
            valid_type=Dict,
            help=
            'Output node of second single impurity calculation. See help string of `impurity1_output_node` for more details.'
        )

        spec.input(
            'offset_imp2',
            valid_type=Dict,
            required=True,
            help="""Offset of the secon impurity with respect to the first impurity.
Can be given either via the 'vector' or the 'index' keys in the dictionary.
The 'vector' option allows to give the offset vector in cartesian units and
the 'index' option allows to five the offset vector in units of the lattice
vectors of the host system's structure."""
        )
        spec.input(
            'wf_parameters_overwrite',
            valid_type=Dict,
            required=False,
            help='To add or edit wf_parameters in scf namespace and add run optioins, if needed'
        )
        spec.input(
            'gf_host_remote',
            valid_type=RemoteData,
            required=
            False,  #TODO Add validator that makes sure this is not given together with the host_gf sub-workflow namespace
            help="""RemoteData node of pre-calculated host Green function (i.e. with kkr_flex_wc).
If given then the writeout step of the host GF is omitted."""
        )

        # structure of the workflow
        spec.outline(
            cls.start,  # initialize workflow (set things in context and some consistency checks)
            cls.combine_single_single,  # check the inputs given node whether from single impurity wc.
            cls.create_big_cluster,  # Create the big cluster for both single-single and multi-single
            cls.update_params,  # update wf_parameters of kkr_imp_sub_wc, kkr_flex_wc and run_options
            if_(cls.need_gf_run)(  # check if GF was given in input and can be reused
                cls.run_gf_writeout
            ),  # write out the host GF
            cls.check_host_gf,  # update the host GF
            cls.create_big_potential,  # combine preconverged potentials to big one
            cls.run_kkrimp_scf,  # run the kkrimp_sub workflow to converge the host-imp startpot
            if_(cls.run_jij)(  # Check Jij step should run or not
                cls.run_jij_step
            ),  # run jij step
            cls.return_results  # check if the calculation was successful and return the result nodes
        )

        # define the possible exit codes
        spec.exit_code(999, 'ERROR_SOMETHING_WENT_WRONG', message='ERROR: take a look')
        spec.exit_code(
            900,
            'ERROR_HOST_STRUCTURES_INCONSISTENT',
            message='Host structures of impurity 1 and impurity 2 are not identical!'
        )
        spec.exit_code(
            800, 'ERROR_INPUT_NOT_SINGLE_IMP_CALC', message='Impurity input is not a single impurity calculation.'
        )
        spec.exit_code(
            850,
            'ERROR_INPLANE_NEIGHBOR_TOO_SMALL',
            message='i_neighbor_inplane needs to be positive and bigger than 0 for in-plane neighbors'
        )
        spec.exit_code(
            950, 'ERROR_INCONSISTENT_NSPIN_VALUES', message='The impurity calculations have different NSPIN values'
        )
        spec.exit_code(700, 'ERROR_HOST_GF_CALC_FAILED', message='The writeout of the host GF failed')

        #TODO to fix this create_combined_imp_info_cf need to take the different layers into account
        # when the difference vector and the neighbors are created

        # define the outputs of the workflow
        spec.output('workflow_info')
        spec.output('last_calc_output_parameters')
        spec.output('last_potential')
        spec.output('last_calc_remote')
        spec.output('remote_data_gf', required=False)
        spec.output('JijData', required=False)
        spec.output('JijInfo', required=False)

    def start(self):  # pylint: disable=inconsistent-return-statements
        """
        prepare context, run_option, wf_parameter, run_option, wf_parameterss and do some consistency checks
        """
        message = f'INFO: started combine_imps_wc workflow version {self._workflowversion}'
        self.report(message)

        self.ctx.run_options = self._wf_default
        self.ctx.allow_unconverged_inputs = self._wf_default.get('allow_unconverged_inputs', False)
        if 'wf_parameters_overwrite' in self.inputs:
            self.ctx.wf_parameters_overwrite = self.inputs.wf_parameters_overwrite.get_dict()
            self.ctx.allow_unconverged_inputs = self.ctx.wf_parameters_overwrite.get('global', {}).get(
                'allow_unconverged_inputs', False
            )

        exit_code, self.ctx.imp1 = self.get_imp_node_from_input(iimp=1)
        if exit_code != 0:
            return exit_code
        exit_code, self.ctx.imp2 = self.get_imp_node_from_input(iimp=2)
        if exit_code != 0:
            return exit_code
        self.ctx.single_vs_single = True  # It is assumed

        # find and compare host structures for the two imps to make sure the impurities are consistent
        host_structure1 = get_host_structure(self.ctx.imp1)
        host_structure2 = get_host_structure(self.ctx.imp2)
        #TODO this can be relaxed to make sure the same structure is used even if it is not the same node
        if host_structure1.uuid != host_structure2.uuid:
            self.report('host structures inconsistent')
            return self.exit_codes.ERROR_HOST_STRUCTURES_INCONSISTENT  # pylint: disable=maybe-no-member

        # save host structure in context
        self.ctx.host_structure = host_structure1

        # settings for offset between imps
        self.ctx.i_neighbor_inplane = self.inputs.offset_imp2['index']

        # preserve the inputs from scf namespace to context
        self.ctx.scf_kkrimp = self.inputs.scf.kkrimp
        if 'options' in self.inputs.scf:
            self.ctx.scf_options = self.inputs.scf.options
        if 'wf_parameters' in self.inputs.scf:
            self.ctx.scf_wf_parameters = self.inputs.scf.wf_parameters
        # Add some run option here
        self.ctx.jij_option = False
        self.ctx.offset_imp2 = self.inputs.offset_imp2

        # TODO: PRESERVE THE INPUTS FROM host_gf NAMESPACE TO CONTEXT
        # TODO: ALSO EDIT THE RUN_GF_WRITEOUT() FOR THIS CORRESPONDING CHANGES

    # To check the inputs both impurity1_output_node impurity2_output_node from the single kkrimp calc or wf, or combine and single calc or wf.
    def combine_single_single(self):
        """
            This function checks whether the impurity1_output_node and the impurity2_output_node are from (kkr_imp_wc, kkr_imp_wc) or (combine_imps_wc, kkr_imp_wc), and always keeps the combine_imps_wc as the first impurity node for combine cluster.
        """
        single_single = True

        single_imp_1 = True
        single_imp_2 = True

        imp_1 = self.ctx.imp1
        imp_2 = self.ctx.imp2
        # check for the impurity1 whether from single kkr_imp_wc or not
        if imp_1.process_class == KkrimpCalculation:
            try:
                impurity_info = imp_1.inputs.impurity_info
            except:
                host_GF = imp_1.inputs.host_Greenfunction_folder
                host_GF_calc = host_GF.base.links.get_incoming(node_class=CalcJobNode).first().node
                impurity_info = host_GF_calc.inputs.impurity_info
            Zimp_num_1 = impurity_info.get_dict().get('Zimp')

            if isinstance(Zimp_num_1, list):
                if len(Zimp_num_1) > 1:
                    single_imp_1 = False

        elif imp_1.process_class == kkr_imp_sub_wc:
            combine_wc = imp_1.get_incoming(node_class=combine_imps_wc).all()
            Zimp_num_1 = imp_1.inputs.impurity_info.get_dict().get('Zimp')
            if len(Zimp_num_1) > 1:
                single_imp_1 = False
        elif imp_1.process_class == combine_imps_wc:
            single_imp_1 = False

        # check for the impurity2 whether from single kkr_imp_wc or not
        if imp_2.process_class == KkrimpCalculation:
            try:
                impurity_info = imp_2.inputs.impurity_info
            except:
                host_GF = imp_2.inputs.host_Greenfunction_folder
                host_GF_calc = host_GF.base.links.get_incoming(node_class=CalcJobNode).first().node
                impurity_info = host_GF_calc.inputs.impurity_info
            Zimp_num_2 = impurity_info.get_dict().get('Zimp')

            if isinstance(Zimp_num_2, list):
                if len(Zimp_num_2) > 1:
                    single_imp_2 = False
        elif imp_2.process_class == kkr_imp_sub_wc:
            combine_wc = imp_2.get_incoming(node_class=combine_imps_wc).all()
            Zimp_num_2 = imp_2.inputs.impurity_info.get_dict().get('Zimp')
            if len(Zimp_num_2) > 1:
                single_imp_2 = False
        elif imp_2.process_class == combine_imps_wc:
            single_imp_2 = False

        if not (single_imp_1 and single_imp_2):
            single_single = False
            if single_imp_2 == False:
                if single_imp_1 == False:
                    self.report(
                        f"ERROR: Both 'impurity1_output_node' {self.inputs.impurity1_output_node} and 'impurity2_output_node {self.inputs.impurity2_output_node} are from combine_imps_wc."
                    )
                    return self.exit_codes.ERROR_SOMETHING_WENT_WRONG  # pylint: disable=no-member
                else:
                    self.ctx.imp1 = imp_2
                    self.ctx.imp2 = imp_1
        # TODO: Delete this print line for created for deguging
        self.ctx.single_vs_single = single_single
        self.report(
            f'DEBUG: The self.combine_single_single() is successfully done. This is the single-sigle: {self.ctx.single_vs_single}'
        )

    def extract_imps_info_exact_cluster(self):
        """
                This method collects the all exist impurity info as in the exact crystal rather than in the crystal centering the first impurity at (0,0,0) position. Returns the imps_info_in_exact_cluster dict.
        """
        if self.ctx.single_vs_single:
            #TODO what is self.ctx.imp1 i.e self.ctx.imp1==combine_imps_wc for single
            imps_info_in_exact_cluster = self.imps_info_exact_cluster_2imps(
                self.ctx.imp1, self.ctx.imp2, self.inputs.offset_imp2
            )
            self.report(
                f'DEBUG: The is the imps_info_in_exact_cluster dict for single-single imp calc: {imps_info_in_exact_cluster}\n'
            )
            return 0, imps_info_in_exact_cluster
        else:
            imp1_input = self.ctx.imp1
            # This 'if clause' to extract the imps_info_in_exact_cluster from  workflow_info of the input impurity node
            if imp1_input.process_class == combine_imps_wc:
                parent_combine_wc = imp1_input
                out_workflow_info = parent_combine_wc.outputs.workflow_info

            elif imp1_input.process_class == KkrimpCalculation:
                kkrimp_sub = imp1_input.get_incoming(node_class=kkr_imp_sub_wc).all()[0].node
                parent_combine_wc = kkrimp_sub.get_incoming(node_class=combine_imps_wc).all()[0].node
                out_workflow_info = imp1_input.outputs.output_parameters

            elif imp1_input.process_class == kkr_imp_sub_wc:
                parent_combine_wc = imp1_input.get_incoming(node_class=combine_imps_wc).all()[0].node
                out_workflow_info = imp1_input.outputs.workflow_info

            imp2_impurity_info = self.ctx.imp2.inputs.impurity_info

            try:
                imps_info_in_exact_cluster = out_workflow_info.get_dict()['imps_info_in_exact_cluster']  # pylint: disable=possibly-used-before-assignment
            except KeyError:
                parent_input_imp1 = parent_combine_wc.inputs.impurity1_output_node  # TODO: rename combine_wc to the parent_combine_wc
                parent_input_imp2 = parent_combine_wc.inputs.impurity2_output_node
                parent_input_offset = parent_combine_wc.inputs.offset_imp2

                exit_code, parent_imp1_wc_or_calc = self.get_imp_node_from_input(impurity_output_node=parent_input_imp1)
                if exit_code != 0:
                    return exit_code, None
                exit_code, parent_imp2_wc_or_calc = self.get_imp_node_from_input(impurity_output_node=parent_input_imp2)
                if exit_code != 0:
                    return exit_code, None

                # Here below imps_info_in_exact_cluster is construct from the inputs impurity1_output_node, and impurity2_output_node as the idea was not calculated in the old version attempt of the combine_imps_wc.

                imps_info_in_exact_cluster = self.imps_info_exact_cluster_2imps(
                    parent_imp1_wc_or_calc, parent_imp2_wc_or_calc, parent_input_offset
                )
            # Now to add the input impurity info and off set of the present combine_imps_wc
            imps_info_in_exact_cluster['offset_imps'].append(self.ctx.offset_imp2['index'])

            Zimp_2 = imp2_impurity_info.get_dict()['Zimp']
            if isinstance(Zimp_2, list):
                imps_info_in_exact_cluster['Zimps'].append(Zimp_2[0])
            else:
                imps_info_in_exact_cluster['Zimps'].append(Zimp_2)

            imps_info_in_exact_cluster['ilayers'].append(imp2_impurity_info.get_dict()['ilayer_center'])
            # TODO: Delete the below print line as it is for debugging
            self.report(f'DEBUG: The is the imps_info_in_exact_cluster dict: {imps_info_in_exact_cluster}\n')
            return 0, imps_info_in_exact_cluster  # return also exit code

    def get_impinfo_from_hostGF(self, imp_calc):
        """
        Extract impurity infor node from the incoming host GF folder
        """
        if 'impurity_info' in imp_calc.inputs:
            return imp_calc.inputs.impurity_info

        if imp_calc.process_class == KkrimpCalculation:
            GF_input = imp_calc.inputs.host_Greenfunction_folder
        elif imp_calc.process_class == kkr_imp_sub_wc:
            GF_input = imp_calc.inputs.remote_data
        elif imp_calc.process_class == kkr_imp_wc:
            GF_input = imp_calc.inputs.remote_data_gf
        else:
            raise ValueError('Unable to get impinfo')

        parent_calc = GF_input.get_incoming(node_class=CalcJobNode).first().node
        impinfo = parent_calc.inputs.impurity_info

        return impinfo

    def imps_info_exact_cluster_2imps(self, single_imp1_wc, single_imp2_wc, offset_imp2):
        """
            This construct a python dict keeping info about two single inpurities with respect to the original host structure e.i. before transforming the center to the first impurity position.
        """
        impinfo1 = self.get_impinfo_from_hostGF(single_imp1_wc)
        impinfo2 = self.get_impinfo_from_hostGF(single_imp2_wc)
        # imp_info_in_exact_cluster keeps the exact data before creating the cluster will help to add more imps later.
        imps_info_in_exact_cluster = {
            'Zimps': [],
            'ilayers': [],
            'offset_imps': [0]
        }  # This number is taking the first imp.
        #offset_imp contains offset_index for imps 2nd, 3rd so on
        zimp_1 = impinfo1.get_dict().get('Zimp')
        zimp_2 = impinfo2.get_dict().get('Zimp')
        if isinstance(zimp_1, list):
            zimp_1 = zimp_1[0]
        if isinstance(zimp_2, list):
            zimp_2 = zimp_2[0]
        imps_info_in_exact_cluster['Zimps'].append(zimp_1)
        imps_info_in_exact_cluster['Zimps'].append(zimp_2)

        imps_info_in_exact_cluster['ilayers'].append(impinfo1.get_dict().get('ilayer_center'))
        imps_info_in_exact_cluster['ilayers'].append(impinfo2.get_dict().get('ilayer_center'))
        imps_info_in_exact_cluster['offset_imps'].append(offset_imp2.get_dict().get('index'))

        return imps_info_in_exact_cluster

    def get_imp_node_from_input(self, impurity_output_node=None, iimp=1):
        """
        extract impurty calculation from impurity_output_node as inputs
        """
        if impurity_output_node == None:
            if iimp == 1:
                imp_out = self.inputs.impurity1_output_node
            else:
                imp_out = self.inputs.impurity2_output_node
        else:
            imp_out = impurity_output_node

        kkrimpcalc_parents = imp_out.get_incoming(node_class=KkrimpCalculation).all()
        if len(kkrimpcalc_parents) > 0:
            imp = kkrimpcalc_parents[0].node
        else:
            inc = imp_out.get_incoming(link_label_filter='workflow_info').all()
            if len(inc) != 1:
                self.report(f'input node of imp {iimp} inconsistent')
                return self.exit_codes.ERROR_INPUT_NODE_INCONSISTENT, None  # pylint: disable=maybe-no-member
            imp = inc[0].node

        # consistency checks of input nodes
        # check if input calc was converged etc.
        if not self._check_input_imp(imp):
            self.report(f'something wrong with imp {iimp}: {imp}')
            return self.exit_codes.ERROR_SOMETHING_WENT_WRONG, None  # pylint: disable=maybe-no-member

        # return exit code and imp calculation
        return 0, imp

    def _check_input_imp(self, imp_calc_or_wf):
        """
        check if input calculation is of the right process_class and if it did converge
        """

        if imp_calc_or_wf.process_class != KkrimpCalculation:
            # if imp_calc_or_wf is not a KkrimpCalculation
            # check if imp_calc_or_wf is kkr_imp_wc, kkr_imp_sub_wc or combine_imps_wc workflow
            if not isinstance(imp_calc_or_wf, WorkChainNode):
                self.report(f'impurity_workflow not a WorkChainNode: {imp_calc_or_wf}')
                return False

            if not (
                imp_calc_or_wf.process_class == kkr_imp_wc or imp_calc_or_wf.process_class == kkr_imp_sub_wc or
                imp_calc_or_wf.process_class == self.__class__
            ):
                self.report(f'impurity_workflow class is wrong: {imp_calc_or_wf}')
                return False

        if (not self.ctx.allow_unconverged_inputs):
            # calculation should be converged
            if imp_calc_or_wf.process_class == KkrimpCalculation:
                if not imp_calc_or_wf.outputs.output_parameters['convergence_group']['calculation_converged']:
                    self.report(f'impurity calculation not converged: {imp_calc_or_wf}')
                    return False
            if imp_calc_or_wf.process_class == kkr_imp_wc:
                if not imp_calc_or_wf.outputs.workflow_info.get_dict().get('converged'):
                    self.report('impurity_workflow not converged')
                    return False
            elif (imp_calc_or_wf.process_class == kkr_imp_sub_wc or imp_calc_or_wf.process_class == self.__class__):
                if not imp_calc_or_wf.outputs.workflow_info.get_dict().get('convergence_reached'):
                    self.report('impurity_workflow not converged')
                    return False

        # all checks passed
        return True

    def create_big_cluster(self):  # pylint: disable=inconsistent-return-statements
        """
        combine imp clusters of the two imps
        """
        imp1 = self.ctx.imp1
        imp2 = self.ctx.imp2
        single_single = self.ctx.single_vs_single
        if single_single:
            if _debug:
                print('DEBUG:', list(imp1.inputs))
            impinfo1 = self.get_impinfo_from_hostGF(imp1)
            impinfo2 = self.get_impinfo_from_hostGF(imp2)
            # impinfo1 = imp1.inputs.impurity_info
        else:
            if imp1.process_class == self.__class__:
                imp1 = imp1.get_outgoing(node_class=kkr_imp_sub_wc).all()[0].node
            impinfo1 = imp1.inputs.impurity_info
            self.report(f'DEBUG: impinfo1 : {impinfo1.get_dict()} .')

            if imp2.process_class == self.__class__:
                imp2 = imp2.get_outgoing(node_class=kkr_imp_sub_wc).all()[0].node
            impinfo2 = imp2.inputs.impurity_info
            self.report(f'DEBUG: impinfo2 : {impinfo2.get_dict()} .')

        host_structure = self.ctx.host_structure
        offset_imp2 = self.inputs.offset_imp2

        self.report('create combined imp_info:')
        self.report(f'host structure: {host_structure}')
        self.report(f'imp info 1: {impinfo1}')
        self.report(f'imp info 2: {impinfo2}')

        exit_code, self.ctx.imps_info_in_exact_cluster = self.extract_imps_info_exact_cluster()
        if exit_code != 0:
            return exit_code

        imps_info_in_exact_cluster = self.ctx.imps_info_in_exact_cluster
        if single_single:
            if offset_imp2.get_dict()['index'] < 0:
                return self.exit_codes.ERROR_INPLANE_NEIGHBOR_TOO_SMALL  # pylint: disable=maybe-no-member
            if 'ilayer_center' in impinfo1.get_dict() and 'ilayer_center' in impinfo2.get_dict():
                if impinfo1['ilayer_center'] == impinfo2['ilayer_center'] and self.inputs.offset_imp2['index'] < 1:
                    return self.exit_codes.ERROR_INPLANE_NEIGHBOR_TOO_SMALL  # pylint: disable=maybe-no-member
        else:
            imp_offset_index = offset_imp2['index']
            imp2_ilayer = impinfo2['ilayer_center']

            for offset, ilayer in zip(
                imps_info_in_exact_cluster['offset_imps'][:-1], imps_info_in_exact_cluster['ilayers'][:-1]
            ):
                if (offset, ilayer) == (imp_offset_index, imp2_ilayer):
                    self.report(
                        f"ERROR: The new impurity is overlaping with the existing impurities. Change the 'ilayer_center': {ilayer} or 'offset_index'{offset}."
                    )
                    return self.exit_codes.ERROR_SOMETHING_WENT_WRONG  # pylint: disable=no-member

        # get zimp of imp1
        if single_single:
            _, is_single_imp = self.get_and_check_zimp_list(impinfo1)
            if not is_single_imp:
                return self.exit_codes.ERROR_INPUT_NOT_SINGLE_IMP_CALC  # pylint: disable=maybe-no-member

        # do the same for imp2
        _, is_single_imp = self.get_and_check_zimp_list(impinfo2)
        if not is_single_imp:
            return self.exit_codes.ERROR_INPUT_NOT_SINGLE_IMP_CALC  # pylint: disable=maybe-no-member
        self.report(f'DEBURG: This is the exact imps: {imps_info_in_exact_cluster}')

        # create combined cluster, offset of second imp is extracted from i_neighbor_inplane
        out_dict = create_combined_imp_info_cf(
            host_structure, impinfo1, impinfo2, offset_imp2, Dict(dict=imps_info_in_exact_cluster), Bool(single_single)
        )

        self.ctx.imp_info_combined = out_dict['imp_info_combined']
        self.ctx.kickout_info = out_dict['kickout_info']
        self.report(
            f"DEBUG: imp_info_combined: {out_dict['imp_info_combined']}\n kickout_info: {out_dict['kickout_info']}"
        )

    def get_and_check_zimp_list(self, impurity_info):
        """
        extract zimp from impurity_info node and check if it is consistent (needs to be single impurity calculation)
        """
        is_single_imp = True

        zimp = get_zimp(impurity_info)

        # check if calculation is single imp calculation
        if len(zimp) != 1:
            is_single_imp = False

        return zimp, is_single_imp

    def need_gf_run(self):
        """
        check if GF was given in input and can be reused (then return Falser which means no gf needs to be calculated)
        """
        if 'gf_host_remote' in self.inputs:
            return False

        return True

    def run_gf_writeout(self):
        """
        Write out the host GF
        """

        # create process builder for gf_writeout workflow
        builder = kkr_flex_wc.get_builder()
        builder.impurity_info = self.ctx.imp_info_combined
        builder.kkr = self.inputs.host_gf.kkr

        if 'wf_parameters' in self.inputs.host_gf:
            self.report(f'set wf_parameters {self.inputs.host_gf.wf_parameters.get_dict()}')
            builder.wf_parameters = self.inputs.host_gf.wf_parameters

        if 'options' in self.inputs.host_gf:
            builder.options = self.inputs.host_gf.options

        if 'params_kkr_overwrite' in self.inputs.host_gf:
            self.report(
                'INFO: using params_kkr_overwrite in host_gf step: {}'.format(
                    self.inputs.host_gf.params_kkr_overwrite.get_dict()
                )
            )
            builder.params_kkr_overwrite = self.inputs.host_gf.params_kkr_overwrite

        # find converged_host_remote input (converged potential of host system)
        gf_writeout_calc = None
        if self.ctx.imp1.process_class == KkrimpCalculation:
            #take gf_writeout directly from input to KkrimpCalculation
            gf_writeout_calc = self.ctx.imp1.inputs.host_Greenfunction_folder.get_incoming(node_class=KkrCalculation
                                                                                           ).first().node
        elif self.ctx.imp1.process_class == kkr_imp_sub_wc:
            imp1_sub = self.ctx.imp1
        else:
            if _debug:
                print('DEBUG:', self.ctx.imp1, list(self.ctx.imp1.inputs))
            imp1_sub = self.ctx.imp1.get_outgoing(node_class=kkr_imp_sub_wc).first().node
        if gf_writeout_calc is None:
            gf_writeout_calc = imp1_sub.inputs.remote_data.get_incoming(node_class=KkrCalculation).first().node  # pylint: disable=possibly-used-before-assignment
        builder.remote_data = gf_writeout_calc.inputs.parent_folder

        # set label and description of the calc
        sub_label = 'GF writeout combined imps'
        sub_description = 'GF writeout sub workflow for combine_imps_wc '
        builder.metadata.label = sub_label  # pylint: disable=no-member
        builder.metadata.description = sub_description  # pylint: disable=no-member

        # now submit the workflow
        future = self.submit(builder)

        self.report(f'INFO: running GF writeout (pk: {future.pk})')

        return ToContext(gf_writeout=future)

    def check_host_gf(self):
        """
        Check if host gf is there
        """
        self.ctx.host_gf_ok = True

        if self.need_gf_run():  # check only if the calculation was run
            if not self.ctx.gf_writeout.is_finished_ok:
                self.ctx.host_gf_ok = False

        #TODO check if input host gf remote is consistent

        if not self.ctx.host_gf_ok:
            return self.exit_codes.ERROR_HOST_GF_CALC_FAILED  # pylint: disable=no-member

    def create_big_potential(self):  # pylint: disable=inconsistent-return-statements
        """
        combine preconverged potentials to big one
        """

        # get data from context
        imp1 = self.ctx.imp1
        imp2 = self.ctx.imp2
        kickout_info = self.ctx.kickout_info

        nspin1, pot_imp1 = get_nspin_and_pot(imp1)
        nspin2, pot_imp2 = get_nspin_and_pot(imp2)

        # check consistency of nspin for the two calculations
        if nspin1 != nspin2:
            return self.exit_codes.ERROR_INCONSISTENT_NSPIN_VALUES  # pylint: disable=maybe-no-member

        # now combine potentials
        output_potential_sfd_node = combine_potentials_cf(kickout_info, pot_imp1, pot_imp2, Int(nspin1))

        self.ctx.combined_potentials = output_potential_sfd_node

    # To collate and combine the wf_parameters_overwrite and scf_wf_parameters
    def update_params(self):
        """
        Update the parameters in scf_wf_parameters according to wf_parameters_overwrite if
        any change occur there and also add the run options.
        """

        scf_wf_parameters = self.ctx.scf_wf_parameters.get_dict()

        if 'wf_parameters' in self.inputs.host_gf:
            wf_parameters_flex = self.inputs.host_gf.wf_parameters.get_dict()
        else:
            wf_parameters_flex = {}
        run_options = self.ctx.run_options
        # Update the scf_wf_parameters from the wf_parameters_overwrite
        if 'wf_parameters_overwrite' in self.inputs:
            wf_parameters_overwrite = self.ctx.wf_parameters_overwrite

            for key, val in wf_parameters_overwrite.items():
                # Update the scf_wf_parameters from wf_parameters_overwrite
                if key in list(scf_wf_parameters.keys()):
                    if wf_parameters_overwrite[key] != scf_wf_parameters[key]:
                        scf_old_val = scf_wf_parameters[key]
                        scf_wf_parameters[key] = val
                        msg = f'INFO: Parameter value of {key} set from {scf_old_val} to {val}'
                        self.report(msg)
                else:
                    scf_wf_parameters[key] = val

        # Update the wf_parameters_flex and run_options from the scf_wf_parameters
        key_list = []
        for key, val in scf_wf_parameters.items():
            if key in list(wf_parameters_flex.keys()) or key in list(run_options.keys()):
                # Here preparing the wf parameters for kkr_flex_wc
                if key in list(wf_parameters_flex.keys()):
                    deflt_val = wf_parameters_flex[key]
                    wf_parameters_flex[key] = scf_wf_parameters.get(key, deflt_val)
                # Here preparing the some run option
                if key in list(run_options.keys()):
                    deflt_val = run_options[key]
                    run_options[key] = scf_wf_parameters.get(key, deflt_val)
                    self.report(f'INFO: Probable run option <{key}> is updated here as <{run_options[key]}>')
                key_list.append(key)
        # Here to remove keys from scf_wf_parameters that are needed only for gf_writeout_step and run_option
        val_list = [scf_wf_parameters.pop(key, None) for key in key_list[:]]

        self.report('INFO: The wf_parameters Dict for kkr_imp_sub_wc is ready.')

        self.ctx.run_options = run_options
        self.ctx.wf_parameters_flex = wf_parameters_flex
        self.ctx.scf_wf_parameters = Dict(scf_wf_parameters)

    def run_kkrimp_scf(self):
        """
        run the kkrimp_sub workflow to converge the host-imp startpot
        """

        # construct process builder for kkrimp scf workflow
        builder = kkr_imp_sub_wc.get_builder()
        builder.metadata.label = 'kkrimp scf combined imps'  # pylint: disable=no-member
        builder.metadata.description = f'scf workflow for combined impurities: {self.ctx.imp1.label}, {self.ctx.imp2.label}'  # pylint: disable=no-member

        # add combined impurity-info and startpot
        builder.impurity_info = self.ctx.imp_info_combined
        builder.host_imp_startpot = self.ctx.combined_potentials

        # add host GF (either calculated or form input)
        if 'gf_host_remote' not in self.inputs:
            gf_remote = self.ctx.gf_writeout.outputs.GF_host_remote
        else:
            gf_remote = self.inputs.gf_host_remote
        builder.remote_data = gf_remote

        # settings from scf namespace
        builder.kkrimp = self.inputs.scf.kkrimp
        if 'options' in self.inputs.scf:
            builder.options = self.inputs.scf.options
        if 'wf_parameters' in self.inputs.scf:
            builder.wf_parameters = self.inputs.scf.wf_parameters
        if 'params_overwrite' in self.inputs.scf:
            builder.params_overwrite = self.inputs.scf.params_overwrite

        # take care of LDA+U settings
        add_ldausettings, settings_LDAU_combined = self.get_ldau_combined()
        self.report(f'add LDA+U settings? {add_ldausettings}')
        if add_ldausettings:
            self.report(f'settings_combined: {settings_LDAU_combined.get_dict()}')
            builder.settings_LDAU = settings_LDAU_combined
        if 'settings_LDAU' in self.inputs.scf:
            self.report('Overwriting settings_LDAU_combined with provided settings_LDAU inputs')
            builder.settings_LDAU = self.inputs.scf.settings_LDAU

        # now submit workflow
        future = self.submit(builder)

        self.report(f'INFO: running kkrimp scf workflow for combined impts (uuid= {future.uuid})')

        return ToContext(kkrimp_scf_sub=future)

    def run_jij(self):
        if not self.ctx.kkrimp_scf_sub.is_finished_ok:
            msg = self.exit_codes.ERROR_SOMETHING_WENT_WRONG  # pylint: disable=no-member
            self.report(msg)
            return False
        # TODO : work here from the RUNOPT instead of scf_wf_parameters, because all the RUNOPT and wf_parameters_flex are updated in the update parameter funcion.
        run_options = self.ctx.run_options
        if 'jij_run' in list(run_options.keys()):
            self.ctx.jij_option = run_options['jij_run']

        return self.ctx.jij_option

    # To launch jij step
    def run_jij_step(self):
        """Run the jij calculation with the converged combined_imp_host_pot"""
        if self.ctx.kkrimp_scf_sub.is_finished_ok:
            msg = 'INFO: kkr_imp_sub_wc for combined impurity cluster is succefully done and jij step is getting prepared for launching.'
            self.report(msg)
            combined_scf_node = self.ctx.kkrimp_scf_sub
        else:
            return self.exit_codes.ERROR_SOMETHING_WENT_WRONG  # pylint: disable=no-member

        last_calc = load_node(combined_scf_node.outputs.workflow_info['last_calc_nodeinfo']['uuid'])
        builder = last_calc.get_builder_restart()

        builder.pop('parent_calc_folder')
        builder.impurity_potential = combined_scf_node.outputs.host_imp_pot
        param_dict = {k: v for k, v in builder.parameters.get_dict().items() if v is not None}
        param_dict['CALCJIJMAT'] = 1  # activate Jij calculation, leave the rest as is

        builder.parameters = Dict(param_dict)
        builder.metadata.label = 'KKRimp_Jij (' + last_calc.label.split('=')[1][3:]

        future = self.submit(builder)
        self.report(f'INFO: Submiting Jij calculation (uuid={future.uuid})')

        return ToContext(imp_scf_combined_jij=future)

    def get_ldau_combined(self):
        """
        check if impurity input calculations have LDA+U settings in input and add this here if needed
        """

        if self.ctx.imp1.process_class == KkrimpCalculation or self.ctx.imp1.process_class == kkr_imp_sub_wc:
            imp1_has_ldau = 'settings_LDAU' in self.ctx.imp1.inputs
            if imp1_has_ldau:
                settings_LDAU1 = self.ctx.imp1.inputs.settings_LDAU
                self.report('found LDA+U settings for imp1')
                if self.ctx.imp1.process_class == KkrimpCalculation:
                    retrieved1 = self.ctx.imp1.outputs.retrieved
                elif self.ctx.imp1.process_class == kkr_imp_sub_wc:
                    retrieved1 = self.ctx.imp1.get_outgoing(node_class=KkrimpCalculation).first().node.outputs.retrieved
        elif self.ctx.imp1.process_class == combine_imps_wc:
            imp1_has_ldau = 'settings_LDAU' in self.ctx.imp1.get_outgoing(node_class=kkr_imp_sub_wc).first().node.inputs
            if imp1_has_ldau:
                settings_LDAU1 = self.ctx.imp1.get_outgoing(node_class=kkr_imp_sub_wc).first().node.inputs.settings_LDAU
                self.report('found LDA+U settings for imp1')
                retrieved1 = self.ctx.imp1.get_outgoing(node_class=kkr_imp_sub_wc
                                                        ).first().node.get_outgoing(node_class=KkrimpCalculation
                                                                                    ).first().node.outputs.retrieved

        if self.ctx.imp2.process_class == KkrimpCalculation or self.ctx.imp2.process_class == kkr_imp_sub_wc:
            imp2_has_ldau = 'settings_LDAU' in self.ctx.imp2.inputs
            if imp2_has_ldau:
                settings_LDAU2 = self.ctx.imp2.inputs.settings_LDAU
                self.report('found LDA+U settings for imp2')
                if self.ctx.imp2.process_class == KkrimpCalculation:
                    retrieved2 = self.ctx.imp2.outputs.retrieved
                elif self.ctx.imp2.process_class == kkr_imp_sub_wc:
                    retrieved2 = self.ctx.imp2.get_outgoing(node_class=KkrimpCalculation).first().node.outputs.retrieved
        elif self.ctx.imp2.process_class == combine_imps_wc:
            imp2_has_ldau = 'settings_LDAU' in self.ctx.imp2.get_outgoing(node_class=kkr_imp_sub_wc).first().node.inputs
            if imp2_has_ldau:
                settings_LDAU2 = self.ctx.imp2.get_outgoing(node_class=kkr_imp_sub_wc).first().node.inputs.settings_LDAU
                self.report('found LDA+U settings for imp2')
                retrieved2 = self.ctx.imp2.get_outgoing(node_class=kkr_imp_sub_wc
                                                        ).first().node.get_outgoing(node_class=KkrimpCalculation
                                                                                    ).first().node.outputs.retrieved

        if imp1_has_ldau and imp2_has_ldau:
            # combine LDA+U settings of the two imps
            settings_LDAU_combined = combine_settings_ldau(
                settings_LDAU1=settings_LDAU1,  # pylint: disable=used-before-assignment
                retrieved1=retrieved1,  # pylint: disable=used-before-assignment
                settings_LDAU2=settings_LDAU2,  # pylint: disable=used-before-assignment
                retrieved2=retrieved2,  # pylint: disable=used-before-assignment
                kickout_info=self.ctx.kickout_info
            )
        elif imp1_has_ldau:
            # use only LDA+U settings of imp 1
            settings_LDAU_combined = combine_settings_ldau(
                settings_LDAU1=settings_LDAU1, retrieved1=retrieved1, kickout_info=self.ctx.kickout_info
            )
        elif imp2_has_ldau:
            # add offset to atom index for combined LDA+U settings
            settings_LDAU_combined = combine_settings_ldau(
                settings_LDAU2=settings_LDAU2, retrieved2=retrieved2, kickout_info=self.ctx.kickout_info
            )
        else:
            # return builder unchanged if none of the impurt calculations has LDA+U settings
            return False, {}

        # now add settings_LDAU input to builder
        self.report(
            f'add combined LDAU settings (uuid={settings_LDAU_combined.uuid}): {settings_LDAU_combined.get_dict()}'
        )

        return True, settings_LDAU_combined

    def return_results(self):
        """
        check if the calculation was successful and return the result nodes
        """

        import numpy as np

        self.report('INFO: Return_results:INFO: Return_results:  To collect the WF info and Other results')

        if not self.ctx.kkrimp_scf_sub.is_finished_ok:
            self.report('ERROR: kkrimp convergence step is not finished successfully')
            return self.exit_codes.ERROR_SOMETHING_WENT_WRONG  # pylint: disable=no-member

        # collect results of kkrimp_scf sub-workflow
        kkrimp_scf_sub = self.ctx.kkrimp_scf_sub
        results_kkrimp_sub = kkrimp_scf_sub.outputs.workflow_info
        last_calc = load_node(results_kkrimp_sub['last_calc_nodeinfo']['uuid'])
        output_parameters = last_calc.outputs.output_parameters
        last_remote = last_calc.outputs.remote_folder
        last_pot = kkrimp_scf_sub.outputs.host_imp_pot
        out_dict = {}
        out_dict['workflow_name'] = self.__class__.__name__
        out_dict['workflow_version'] = self._workflowversion

        # collect info of sub workflow
        out_dict['sub_workflows'] = {'kkrimp_scf': {'pk': kkrimp_scf_sub.pk, 'uuid': kkrimp_scf_sub.uuid}}

        # collect some results from scf sub-workflow
        for key in ['successful', 'convergence_value', 'convergence_reached', 'convergence_values_all_steps']:
            out_dict[key] = results_kkrimp_sub[key]

        try:
            magmom_all = np.array(output_parameters['magnetism_group']['spin_moment_per_atom'], dtype=float)[:, -1]
            out_dict['magmoms'] = magmom_all
        except KeyError:
            self.report("No 'magmons' available in 'workflow_info'. Probably, 'mag_init' set to False in the settings.")

        # Parse_jij and collect some info
        is_jij_exist = self.ctx.jij_option
        if is_jij_exist:
            impurity1_output_node = self.inputs.impurity1_output_node
            impurity2_output_node = self.inputs.impurity2_output_node
            jij_calc = self.ctx.imp_scf_combined_jij
            jij_retrieved = jij_calc.outputs.retrieved
            impurity_info = kkrimp_scf_sub.inputs.impurity_info
            out_dict['run_option_info'] = {
                'jij_calc': {
                    'pk': jij_calc.pk,
                    'uuid': jij_calc.uuid,
                    'is_finished_ok': jij_calc.is_finished_ok
                }
            }
            jij_parsed_dict = parse_Jij(jij_retrieved, impurity_info, impurity1_output_node, impurity2_output_node)
        # collect outputs of host_gf sub_workflow if it was done
        if 'gf_host_remote' not in self.inputs:
            gf_writeout = self.ctx.gf_writeout
            gf_sub_remote = gf_writeout.outputs.GF_host_remote

            # add info about sub-workflow to dict output
            out_dict['sub_workflows']['host_gf'] = {'pk': gf_writeout.pk, 'uuid': gf_writeout.uuid}
            # add as output node
            self.out('remote_data_gf', gf_sub_remote)

        # add information on combined cluster and potential
        out_dict['imp_info_combined'] = self.ctx.imp_info_combined.get_dict()
        out_dict['potential_kickout_info'] = self.ctx.kickout_info.get_dict()
        out_dict['imps_info_in_exact_cluster'] = self.ctx.imps_info_in_exact_cluster

        # create results node with input links
        # TODO: Add the imp_scf_combined_jij to link the run_jij_step output on provanace graph
        link_nodes = {'kkrimp_scf_results': results_kkrimp_sub}
        if 'gf_host_remote' in self.inputs:
            link_nodes['GF_host_remote'] = self.inputs.gf_host_remote
        if is_jij_exist:
            link_nodes['Jij_retrieved'] = jij_retrieved
        outputnode = create_out_dict_node(Dict(dict=out_dict), **link_nodes)
        outputnode.label = 'combine_imps_wc_results'

        # add output nodes
        self.out('workflow_info', outputnode)
        self.out('last_potential', last_pot)
        self.out('last_calc_remote', last_remote)
        self.out('last_calc_output_parameters', output_parameters)
        if is_jij_exist:
            self.out('JijData', jij_parsed_dict['Jijdata'])
            self.out('JijInfo', jij_parsed_dict['info'])


@calcfunction
def parse_Jij(retrieved, impurity_info, impurity1_output_node, impurity2_output_node):
    """parser output of Jij calculation and return as ArrayData node"""

    # Collect the zimp for impurity_output_node
    try:
        imp1_z = impurity1_output_node.get_incoming(node_class=kkr_imp_wc
                                                    ).first().node.inputs.impurity_info.get_dict()['Zimp']
    except AttributeError:
        try:
            impurity1_output_node_combine = impurity1_output_node.get_incoming(node_class=combine_imps_wc).first().node
            imp1_z = impurity1_output_node_combine.get_outgoing(node_class=kkr_imp_sub_wc
                                                                ).first().node.inputs.impurity_info.get_dict()['Zimp']
        except AttributeError:
            try:
                imp1_z = impurity1_output_node.get_incoming(node_class=kkr_imp_sub_wc
                                                            ).first().node.inputs.impurity_info.get_dict()['Zimp']
            except AttributeError:
                imp1_z = impurity1_output_node.get_incoming(node_class=KkrimpCalculation
                                                            ).first().node.inputs.impurity_info.get_dict()['Zimp']

    try:
        imp2_z = impurity2_output_node.get_incoming(node_class=kkr_imp_wc
                                                    ).first().node.inputs.impurity_info.get_dict()['Zimp']
    except AttributeError:
        try:
            impurity2_output_node_combine = impurity2_output_node.get_incoming(node_class=combine_imps_wc).first().node
            imp2_z = impurity2_output_node_combine.get_outgoing(node_class=kkr_imp_sub_wc
                                                                ).first().node.inputs.impurity_info.get_dict()['Zimp']
        except AttributeError:
            try:
                imp2_z = impurity2_output_node.get_incoming(node_class=kkr_imp_sub_wc
                                                            ).first().node.inputs.impurity_info.get_dict()['Zimp']
            except AttributeError:
                imp2_z = impurity2_output_node.get_incoming(node_class=KkrimpCalculation
                                                            ).first().node.inputs.impurity_info.get_dict()['Zimp']

    with retrieved.open('out_Jijmatrix') as _f:
        jijdata = np.loadtxt(_f)

    impurity_info = impurity_info.get_dict()
    pos = np.array(impurity_info['imp_cls'])
    z = np.array(impurity_info['imp_cls'])[:, 4]

    if not isinstance(imp1_z, list):
        Vpos = list(np.where(z == imp1_z)[0]) + list(np.where(z == imp2_z)[0])
    else:
        Vpos = [np.where(z == i)[0][0] for i in imp1_z] + list(np.where(z == imp2_z)[0])
    Ry2eV = get_Ry2eV()

    # extract number of atoms
    natom = int(np.sqrt(jijdata.shape[0] / 3 / 3))

    # reshape data
    jij_reshape = jijdata.reshape(3, natom, natom, 3, 3)  # iter, i, j, k, l (Jij_k,l matrix)

    # now combine iterations to get full 3 by 3 Jij matrices for all atom pairs
    jij_combined_iter = np.zeros((natom, natom, 3, 3))
    for iatom in range(natom - 1):
        for jatom in range(natom)[iatom + 1:]:
            for iiter in range(3):
                if iiter == 0:
                    # first iteration with theta, phi = 0, 0
                    # take complete upper block from here since this calculation should be converged best
                    # (rotated moments only one-shot calculations)
                    jij_combined_iter[iatom, jatom, 0, 0] = jij_reshape[iiter, iatom, jatom, 0, 0]
                    jij_combined_iter[iatom, jatom, 0, 1] = jij_reshape[iiter, iatom, jatom, 0, 1]
                    jij_combined_iter[iatom, jatom, 1, 0] = jij_reshape[iiter, iatom, jatom, 1, 0]
                    jij_combined_iter[iatom, jatom, 1, 1] = jij_reshape[iiter, iatom, jatom, 1, 1]
                elif iiter == 1:
                    # second iteraton with theta, phi = 90, 0
                    jij_combined_iter[iatom, jatom, 1, 2] = jij_reshape[iiter, iatom, jatom, 1, 2]
                    jij_combined_iter[iatom, jatom, 2, 1] = jij_reshape[iiter, iatom, jatom, 2, 1]
                    jij_combined_iter[iatom, jatom, 2, 2] = jij_reshape[iiter, iatom, jatom, 2, 2]
                else:
                    # from third iteration with theta, phi = 90, 90
                    jij_combined_iter[iatom, jatom, 0, 2] = jij_reshape[iiter, iatom, jatom, 0, 2]
                    jij_combined_iter[iatom, jatom, 2, 0] = jij_reshape[iiter, iatom, jatom, 2, 0]
                    # add this value to z-z component and average
                    jij_combined_iter[iatom, jatom, 2, 2] += jij_reshape[iiter, iatom, jatom, 2, 2]
                    jij_combined_iter[iatom, jatom, 2, 2] *= 0.5

    # finally convert to meV units (and sign change to have positive number indicate ferromagnetism and negative number antiferromagnetism)
    jij_combined_iter *= -1. * Ry2eV * 1000

    jij_trace = (jij_combined_iter[:, :, 0, 0] + jij_combined_iter[:, :, 1, 1] + jij_combined_iter[:, :, 2, 2]) / 3
    Dij_vec = np.array([(jij_combined_iter[:, :, 1, 2] - jij_combined_iter[:, :, 2, 1]),
                        (jij_combined_iter[:, :, 2, 0] - jij_combined_iter[:, :, 0, 2]),
                        (jij_combined_iter[:, :, 0, 1] - jij_combined_iter[:, :, 1, 0])])

    plotdata = []

    #return jij_combined_iter
    out_txt = 'Output Jij values between V impurities:\n i          j           Jij (meV)           Dij(meV)            D/J             i_zimp          j_zimp \n---------------------------------------------------------------------------\n'
    for iatom in range(natom - 1):
        for jatom in range(natom)[iatom + 1:]:
            if iatom != jatom and iatom in Vpos and jatom in Vpos:
                J = jij_trace[iatom, jatom]
                Dx, Dy, Dz = Dij_vec[0, iatom, jatom], Dij_vec[1, iatom, jatom], Dij_vec[2, iatom, jatom]
                D = np.sqrt(Dx**2 + Dy**2 + Dz**2)
                out_txt += '%3i         %3i         %15.5e          %15.5e          %15.5e          %4i         %4i\n' % (
                    iatom, jatom, J, D, D / J, pos[iatom][4], pos[jatom][4]
                )
                rdiff = pos[jatom] - pos[iatom]
                plotdata.append([rdiff[0], rdiff[1], rdiff[2], J, D, Dx, Dy, Dz])
    plotdata = np.array(plotdata)

    a = ArrayData()
    a.set_array('JijData', plotdata)

    return {'Jijdata': a, 'info': Dict({'text': out_txt})}



# -*- coding: utf-8 -*-
"""
This module contains tools used in LDA+U calculations
"""


def get_ldaupot_text(ldau_settings, ef_Ry, natom, initialize=True, return_luj=False):
    """
    create the text for the ldaupot file
    """
    from masci_tools.io.common_functions import get_Ry2eV

    eV2Ry = 1. / get_Ry2eV()

    # these lists are extracted from ldau_settings node and then written to the ldaupot file
    iatoms_ldau = []
    lopt = []
    ueff = []
    jeff = []
    eref = []

    # extract values from ldau_settings
    for key, val in ldau_settings.items():
        if 'iatom' in key:
            iatoms_ldau.append(int(key.split('=')[1]))
            lopt.append(val['L'])
            # add values in Ry units
            jeff.append(val['J'] * eV2Ry)
            ueff.append(val['U'] * eV2Ry)
            eref.append(val['Eref_EF'] * eV2Ry + ef_Ry)

    if initialize:
        # this means we initialize this file
        ldaurun = 0
    else:
        # this means wldau etc are reused (need to be added to the file)
        ldaurun = 1

    # collect text which is written to ldaupot
    txt = [f'{ldaurun} ']
    txt_lopt, txt_jeff, txt_ueff, txt_eref = [], [], [], []
    ii = 0
    for iatom in range(natom):
        if iatom not in iatoms_ldau:
            txt_lopt += [f'{-1} ']
            txt_jeff += [f'{0.0} ']
            txt_ueff += [f'{0.0} ']
            txt_eref += [f'{0.0} ']
        else:
            txt_lopt += [f'{lopt[ii]} ']
            txt_jeff += [f'{jeff[ii]} ']
            txt_ueff += [f'{ueff[ii]} ']
            txt_eref += [f'{eref[ii]} ']
            ii += 1
    txt += ['\n'] + txt_lopt + ['\n'] + txt_ueff + ['\n'] + txt_jeff + ['\n'] + txt_eref
    txt += ['\nwldau\nuldau\nphi\n']

    # add initial matrices
    if initialize and 'initial_matrices' in ldau_settings.keys():
        # save for consistency check
        nldauatoms = ii

        # change first number from 0 to 1 to signal reading-in instead of calculating initial matrices
        txt[0] = '1 '

        # remove last dummy line, will be replaced with starting values now
        txt[-1] = '\n'
        txt_wldau = ['wldau\n']
        txt_uldau = ['uldau\n']
        txt_phi = ['phi\n']
        ii = 0
        for iatom in range(natom):
            if iatom in iatoms_ldau:
                txt_wldau += [f'atom {iatom+1}\n'] + ldau_settings['initial_matrices'][f'iatom={iatom}']['wldau']
                txt_uldau += [f'atom {iatom+1}\n'] + ldau_settings['initial_matrices'][f'iatom={iatom}']['uldau']
                txt_phi += [f'atom {iatom+1}\n'] + ldau_settings['initial_matrices'][f'iatom={iatom}']['phi']
                ii += 1  # count number of atoms

        # consistency check
        if nldauatoms != ii:
            raise ValueError('initial_matrices input inconsistent')

        # add additional lines to txt
        txt = txt + txt_wldau + txt_uldau + txt_phi

    if not return_luj:
        return txt
    else:
        return txt, lopt, jeff, ueff, eref, iatoms_ldau


def get_ldaumatrices(retrieved):
    """
    Take retrieved folder of KkrimpCalculation and extract ldaupot file
    If it exists we extract the LDAU matrices 'wldau', 'uldau' and 'phi'
    """
    from aiida.common.folders import SandboxFolder
    from aiida_kkr.calculations import KkrimpCalculation

    has_ldaupot_file = False
    txt_dict_ldaumats = {}

    # create Sandbox to extract ldaupot file there
    with SandboxFolder() as tempfolder:
        # extract ldaupot file to tempfolder if it exists
        has_ldaupot_file = KkrimpCalculation.get_ldaupot_from_retrieved(retrieved, tempfolder)
        # now read ldau matrices and store in txt_dict_ldaumats
        if has_ldaupot_file:
            with tempfolder.open(KkrimpCalculation._LDAUPOT + '_old') as ldaupot_file:
                # read file and look for keywords to identify where matrices are stored in the file
                txt = ldaupot_file.readlines()
                ii = 0
                for line in txt:
                    if 'wldau' in line:
                        iwldau = ii
                    if 'uldau' in line:
                        iuldau = ii
                    if 'phi' in line:
                        iphi = ii
                    ii += 1
                # save matrices to output dict
                txt_dict_ldaumats['wldau'] = txt[iwldau + 1:iuldau]  # pylint: disable=possibly-used-before-assignment
                txt_dict_ldaumats['uldau'] = txt[iuldau + 1:iphi]  # pylint: disable=possibly-used-before-assignment
                txt_dict_ldaumats['phi'] = txt[iphi + 1:]  # pylint: disable=possibly-used-before-assignment

    return has_ldaupot_file, txt_dict_ldaumats


def get_LDAU_initmatrices_dict(txts_ldaumat1, offset=0):
    """
    Put the initial matrices text content of `get_ldaumatrices` output in a
    Dictionnary form sorted by atoms.
    """

    # Loop through characters ('wldau', 'uldau', 'phi')
    for char in ['wldau', 'uldau', 'phi']:
        atom_line = []
        iatoms = []

        # Save atom number and lines in which atoms information starts
        for iline in range(len(txts_ldaumat1[char])):
            if txts_ldaumat1[char][iline][:4] == 'atom':
                iatom = txts_ldaumat1[char][iline].split()[1]
                iatoms.append(iatom)
                atom_line.append(iline)
        atom_line.append(len(txts_ldaumat1[char]))

        mat = []
        # Extract matrices based on atom lines
        for iatom in range(len(atom_line) - 1):
            mat.append(txts_ldaumat1[char][atom_line[iatom] + 1:atom_line[iatom + 1]])

        # Assign matrices to corresponding variables based on the character
        if char == 'wldau':
            wldaumat = mat
        elif char == 'uldau':
            uldaumat = mat
        elif char == 'phi':
            phimat = mat

    # Create a dictionary to store the organized data
    LDAU_initmatrices_dict = {}

    # Fill the dictionary with atom-wise information
    for iatom in range(len(iatoms)):
        LDAU_initmatrices_dict[f'iatom={int(iatoms[iatom])-1+offset}'] = {}
        LDAU_initmatrices_dict[f'iatom={int(iatoms[iatom])-1+offset}']['wldau'] = wldaumat[iatom]  # pylint: disable=possibly-used-before-assignment
        LDAU_initmatrices_dict[f'iatom={int(iatoms[iatom])-1+offset}']['uldau'] = uldaumat[iatom]  # pylint: disable=used-before-assignment
        LDAU_initmatrices_dict[f'iatom={int(iatoms[iatom])-1+offset}']['phi'] = phimat[iatom]  # pylint: disable=used-before-assignment

    return LDAU_initmatrices_dict



# -*- coding: utf-8 -*-
"""
This module contains helper functions and tools for manipulating the impurity info node
"""

import numpy as np
from aiida.engine import calcfunction
from aiida.orm import Dict
from aiida_kkr.tools.tools_kkrimp import create_scoef_array
from masci_tools.io.common_functions import get_alat_from_bravais

__copyright__ = (u'Copyright (c), 2023, Forschungszentrum Jülich GmbH, '
                 'IAS-1/PGI-1, Germany. All rights reserved.')
__license__ = 'MIT license, see LICENSE.txt file'
__version__ = '0.1.0'
__contributors__ = (u'Philipp Rüßmann')

# activate debug writeout
debug = False


def get_zimp(impurity_info):
    """
    extract zimp from impurity_info node and return as list
    """
    zimp = impurity_info.get_dict().get('Zimp')

    try:
        zimp = list(zimp)
    except TypeError:
        zimp = [zimp]

    return zimp


def get_scoef_single_imp(host_structure, impinfo_node):
    """
    Create scoef cluster from host_structure and impurity info nodes

    :para impinfo_node: Impurity info node (should have at least Rcut and ilayer_center in dict)
    :type impinfo_node: aiida.orm.Dict
    :para host_structure: structure of host crystal into which impurities are embedded
    :type host_structure: aiida.orm.StructureData

    :return: scoef array (positions [x,y,z], layer index, distance to first position in imp cluster)
    :type: numpy.array
    """
    # check if we can read it from a node extra
    if 'imp_cls' not in impinfo_node.extras:
        impinfo = impinfo_node.get_dict()
        Rcut = impinfo.get('Rcut', None)
        hcut = impinfo.get('hcut', -1.)
        cylinder_orient = impinfo.get('cylinder_orient', [0., 0., 1.])
        ilayer_center = impinfo.get('ilayer_center', 0)

        clust = create_scoef_array(host_structure, Rcut, hcut, cylinder_orient, ilayer_center)
        # sort after distance
        clust = clust[(clust[:, -1]).argsort()]

        # save as extra for second run
        impinfo_node.set_extra('imp_cls', clust)
    else:
        clust = np.array(impinfo_node.extras['imp_cls'])

    return clust


def get_inplane_neighbor(host_structure, i_neighbor, r_out_of_plane):
    """
    create in-plane neighbor
    """
    # if offset of second impurity is given in units of the host structure bravais vectors:
    cell = np.array(host_structure.cell)

    # find and sort list of nearest neighbors
    dist = []
    icount = 0
    N_neighbor_search = 5  # define box in which neighbors are searched
    for j in list(range(N_neighbor_search + 1)
                  ) + list(range(-N_neighbor_search, 0)[::-1]):  # use this to have better ordering (nicer numbers)
        for i in list(range(N_neighbor_search + 1)) + list(range(-N_neighbor_search, 0)[::-1]):
            # create position
            r = cell[0] * i + cell[1] * j + r_out_of_plane
            d = np.sqrt(np.sum(r**2))
            # check if already in list
            add_pair = False
            if icount > 1:
                if abs(np.array(dist)[:, -1] - d).min() > 10**-5:
                    add_pair = True
            else:
                add_pair = True
            # add distance
            if add_pair:
                dist.append([i, j, r[0], r[1], r[2], d])
            icount += 1

    alat = get_alat_from_bravais(cell, host_structure.pbc[2])
    dist = np.array(dist)
    dist[:, 2:] = dist[:, 2:] / alat
    dist_sort_list = dist[:, -1].argsort()

    # find in-plane neighbor
    r_offset = dist[dist_sort_list[i_neighbor]][2:5]  # element 0 is no offset

    return r_offset


def pos_exists_already(pos_list_old, pos_new, debug=False):
    """
    check if pos_new is in pos_list_old
    """
    sort_ref = np.array(pos_list_old)
    dists = np.sqrt(np.sum((sort_ref - np.array(pos_new))**2, axis=1))
    mask = (dists < 10**-5)

    if debug:
        print('check pos', dists.min())

    if dists.min() < 10**-5:
        return True, [i for i in range(len(mask)) if mask[i]]
    else:
        return False, None


def combine_clusters(clust1, clust2_offset, single_single, debug=False):
    """
    combine impurity clusters and remove doubles in there

    :return cluster_combined:
    :return rimp_rel_combined: relative position of impurities in cluster
    :return kickout_list: licst of positions doubled in clust 2
    :return i_neighbor_inplane: position removed from cluster 1 (doubled with impurity position)
    """

    # now combine cluster1 and cluster2
    cluster_combined = list(clust1.copy())

    # check if imp position of cluster 2 is inside and remove that position
    # this ensures that imp2 is not kicked out
    found_double, i_removed_from_1 = pos_exists_already(clust1[:, :3], clust2_offset[0, :3], debug)
    if debug:
        print('i_removed_from_1:', i_removed_from_1, found_double)

    # remove doubled position from impcls1 if there is any
    if i_removed_from_1 is not None:
        #removed = cluster_combined.pop(i_removed_from_1)
        cluster_combined = [clust1[i] for i in range(len(clust1)) if i not in i_removed_from_1]

    # add the rest of the imp cluster of imp 2 if position does not exist already in cluster of imp 1
    kickout_list = []
    for ipos, pos_add in enumerate(clust2_offset):
        if ipos == 0 or not pos_exists_already(np.array(cluster_combined)[:, :3], pos_add[:3], debug)[0]:
            cluster_combined.append(pos_add)
        else:
            kickout_list.append(ipos)
    cluster_combined = np.array(cluster_combined)

    # fix distances for second half
    cluster_combined[:, -1] = np.sqrt(np.sum(cluster_combined[:, :3]**2, axis=1))

    # construct Rimp_rel list
    rimp_rel_combined = [clust1[0, :3]] + [clust2_offset[0, :3]]

    return cluster_combined, rimp_rel_combined, kickout_list, i_removed_from_1


def create_combined_imp_info(
    host_structure, impinfo1, impinfo2, offset_imp2, imps_info_in_exact_cluster, single_single, debug=False
):
    """
    create impurity clusters from impinfo nodes and combine these putting the second
    impurity to the i_neighbor_inplane-th in-plane neighbor
    """
    #TODO: Try it with the self object in the in the input

    imps_info_in_exact_cluster = imps_info_in_exact_cluster.get_dict()
    if single_single:
        zimp1 = imps_info_in_exact_cluster['Zimps'][0]
    zimp2 = imps_info_in_exact_cluster['Zimps'][-1]

    if 'imp_cls' in impinfo1.get_dict():
        clust1 = np.array(impinfo1['imp_cls'])

    elif single_single:
        # create cluster of imp1
        clust1 = get_scoef_single_imp(host_structure, impinfo1)

    # do the same for imp2
    clust2 = get_scoef_single_imp(host_structure, impinfo2)

    # set zimp in scoef file (not used by the code but makes it easier to read the files / debug)
    if single_single:
        clust1[0][4] = zimp1  # pylint: disable=possibly-used-before-assignment
    clust2[0][4] = zimp2
    #if debug:
    #    print('cls1:', clust1)
    #    print('cls2:', clust2)

    if 'r_offset' in offset_imp2.get_dict():
        # use offset given in input
        r_offset = offset_imp2['r_offset']
    else:
        # find offset taking into account the possible out-of-plane vector if the imps are in different layers
        r_out_of_plane = np.array([0, 0, 0])
        center_imp = imps_info_in_exact_cluster['ilayers'][0]
        layer2 = imps_info_in_exact_cluster['ilayers'][-1]
        if center_imp != layer2:
            pos1 = np.array(host_structure.sites[center_imp].position)
            pos2 = np.array(host_structure.sites[layer2].position)
            r_out_of_plane = pos2 - pos1
        i_neighbor_inplane = imps_info_in_exact_cluster['offset_imps'][-1]
        r_offset = get_inplane_neighbor(host_structure, i_neighbor_inplane, r_out_of_plane)
    if debug:
        print('r_offset:', r_offset)

    # add offset to cluster 2
    clust2_offset = clust2.copy()
    clust2_offset[:, :3] += r_offset

    cluster_combined, rimp_rel_combined, kickout_list, i_removed_from_1 = combine_clusters(
        clust1, clust2_offset, single_single, debug
    )

    if 'Rimp_rel' in impinfo1.get_dict():
        rimp_rel_combined = impinfo1['Rimp_rel'] + rimp_rel_combined[1:]

    if debug:
        #print('cls_combined:', cluster_combined)
        print('rimp_rel_combined:', rimp_rel_combined)
        print('kickout_list:', kickout_list)
        print('i_removed_from_1:', i_removed_from_1)

    zimp_combined = imps_info_in_exact_cluster['Zimps']

    # create new imp_info node with imp_cls, Rimp_rel and Zimp definig the cluster and impurity location
    imp_info_combined = Dict({'imp_cls': cluster_combined, 'Zimp': zimp_combined, 'Rimp_rel': rimp_rel_combined})

    # kickout info (used later in cfreation of combined potential)
    kickout_info = Dict({
        'i_removed_from_1': i_removed_from_1,
        'kickout_list': kickout_list,
        'Ncls1': len(clust1),
        'Ncls2': len(clust2),
        'Ncls_combined': len(cluster_combined)
    })

    return {'imp_info_combined': imp_info_combined, 'kickout_info': kickout_info}


@calcfunction
def create_combined_imp_info_cf(
    host_structure, impinfo1, impinfo2, offset_imp2, imps_info_in_exact_cluster, single_single
):
    """
    create impurity clusters from impinfo nodes and combine these putting the second
    impurity to the i_neighbor_inplane-th in-plane neighbor
    """

    return create_combined_imp_info(
        host_structure, impinfo1, impinfo2, offset_imp2, imps_info_in_exact_cluster, single_single
    )



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-KKR package.                                 #
#                                                                             #
# For further information on the license, see the LICENSE.txt file            #
###############################################################################
# pylint: disable=cyclic-import
# ,reimported,unused-import,wrong-import-position
"""
Module with CLI commands for various data types.
"""
import click


@click.group('data')
def cmd_data():
    """Commands to create and inspect data nodes."""


# Import the sub commands to register them with the CLI
from .structure import cmd_structure
from .parameters import cmd_parameter

#TODO
# aiida-kkr data create kkrparams: create a Dict node using the kkrparams class to have help and verification
# aiida-kkr data create options: create options node for workflows (withmpi, resources etc.)
# aiida-kkr data create impinfo: create imputiy info node which has help and verification
# aiida-kkr data create kpoints: create k-points for bandstructure claculations
# aiida-kkr data create noco_angles: create k-points for bandstructure claculations
#also corresponding list and show commands needed
# for list and show we should have additionally imp pot, maybe something else
'''
@cmd_data.command('list')
def cmd_list_kkrstructures():
    """
    Display all KkrstructureData nodes
    """
    from aiida import load_profile
    load_profile()

    from aiida.orm.querybuilder import QueryBuilder
    from aiida.plugins import DataFactory
    KkrStructure = DataFactory('kkr.kkrstructure')

    qb = QueryBuilder()
    qb.append(KkrStructure)
    results = qb.all()

    s = ""
    for result in results:
        obj = result[0]
        s += "{}, pk: {}\n".format(str(obj), obj.pk)
    click.echo(s)
'''



"""
workflows of AiiDA-KKR
"""

# import all workflows here to expose them in `aiida_kkr.workflows` directly
from .voro_start import kkr_startpot_wc
from .dos import kkr_dos_wc, parse_dosfiles
from .bs import kkr_bs_wc, parse_BS_data
from .kkr_scf import kkr_scf_wc
from .eos import kkr_eos_wc, rescale, get_primitive_structure
from .gf_writeout import kkr_flex_wc
from .kkr_imp_sub import kkr_imp_sub_wc
from .kkr_imp import kkr_imp_wc
from .kkr_imp_dos import kkr_imp_dos_wc
from ._combine_imps import combine_imps_wc
from ._decimation import kkr_decimation_wc
from .jijs import kkr_jij_wc
from .imp_BdG import kkrimp_BdG_wc
from .kkr_STM import kkr_STM_wc
